// Extends sbgnviz.elementUtilities
var libs = require("./lib-utilities").getLibs();
var jQuery = ($ = libs.jQuery);

module.exports = function () {
  var options, sbgnvizInstance, elementUtilities, cy;


  function elementUtilitiesExtender(param) {
    sbgnvizInstance = param.sbgnvizInstanceUtilities.getInstance();
    options = param.optionUtilities.getOptions();
    elementUtilities = sbgnvizInstance.elementUtilities;
    cy = param.sbgnvizInstanceUtilities.getCy();

    extend();

    // Return the extended elementUtilities
    return elementUtilities;
  }

  // Extends elementUtilities with chise specific facilities
  function extend() {
    // Section Start
    // Add remove utilities

    elementUtilities.addNode = function (
      x,
      y,
      nodeParams,
      id,
      parent,
      visibility
    ) {
      if (typeof nodeParams != "object") {
        var sbgnclass = nodeParams;
      } else {
        var sbgnclass = nodeParams.class;
        var language = nodeParams.language;
      }

      var css = {};
      //('sbgnclass', sbgnclass)
      // if there is no specific default width or height for
      // sbgnclass these sizes are used
      var defaultWidth = 50;
      var defaultHeight = 50;

      if (visibility) {
        css.visibility = visibility;
      }

      var data = {
        class: sbgnclass,
        language: language,
        bbox: {
          w: defaultWidth,
          h: defaultHeight,
          x: x,
          y: y,
        },
        statesandinfos: [],
        ports: [],
      };

      //console.log("data", data)
      if (id) {
        data.id = id;
      } else {
        data.id = elementUtilities.generateNodeId();
      }

      if (parent) {
        data.parent = parent;
      }

      this.extendNodeDataWithClassDefaults(data, sbgnclass);

      // some defaults are not set by extendNodeDataWithClassDefaults()
      //console.log("sbgnclass", sbgnclass)
      var defaults = this.getDefaultProperties(sbgnclass);

      if (defaults["multimer"]) {
        data.class += " multimer";
      }

      if (defaults["clonemarker"]) {
        data["clonemarker"] = true;
      }
      if (defaults["hypothetical"]) {
        data.class = "hypothetical " + data.class;
      }
      if (defaults["active"]) {
        data.class = "active " + data.class;
      }

      data.bbox["w"] = defaults["width"];
      data.bbox["h"] = defaults["height"];

      //console.log("data", data)
      var eles = cy.add({
        group: "nodes",
        data: data,
        css: css,
        position: {
          x: x,
          y: y,
        },
      });

      //console.log('eles', eles)

      var newNode = eles[eles.length - 1];
      // Get the default ports ordering for the nodes with given sbgnclass
      var ordering = defaults["ports-ordering"];

      // If there is a default ports ordering for the nodes with given sbgnclass and it is different than 'none' set the ports ordering to that ordering
      if (ordering && ordering !== "none") {
        this.setPortsOrdering(newNode, ordering);
      }

      if (
        language == "AF" &&
        !elementUtilities.canHaveMultipleUnitOfInformation(newNode)
      ) {
        if (sbgnclass != "BA plain") {
          // if AF node can have label i.e: not plain biological activity
          var uoi_obj = {
            clazz: "unit of information",
          };
          uoi_obj.label = {
            text: "",
          };

          uoi_obj.bbox = {
            w: 12,
            h: 12,
          };
          elementUtilities.addStateOrInfoBox(newNode, uoi_obj);
        }
      }

      // node bg image was unexpectedly not rendered until it is clicked
      // use this dirty hack until finding a solution to the problem
      var bgImage = newNode.data("background-image");
      //alert('bgImage', bgImage)
      if (bgImage) {
        newNode.data("background-image", bgImage);
      }

      return newNode;
    };

    //Saves old aux units of given node
    elementUtilities.saveUnits = function (node) {
      var tempData = [];
      var index = 0;
      node.data("statesandinfos").forEach(function (ele) {
        tempData.push({
          x: ele.bbox.x,
          y: ele.bbox.y,
          anchorSide: ele.anchorSide,
        });
        index++;
      });
      return tempData;
    };

    //Restores from given data
    elementUtilities.restoreUnits = function (node, data) {
      var index = 0;
      node.data("statesandinfos").forEach(function (ele) {
        if (data !== undefined) {
          ele.bbox.x = data[index].x;
          ele.bbox.y = data[index].y;
          var anchorSide = ele.anchorSide;
          ele.anchorSide = data[index].anchorSide;
          elementUtilities.modifyUnits(node, ele, anchorSide);
          index++;
        }
      });
    };

    //Modify aux unit layouts
    elementUtilities.modifyUnits = function (node, ele, anchorSide) {
      console.log("elementUtilities.modifyUnits", node);
      instance.classes.AuxUnitLayout.modifyUnits(node, ele, anchorSide, cy);
    };

    //For reversible reactions both side of the process can be input/output
    //Group ID identifies to which group of nodes the edge is going to be connected for reversible reactions(0: group 1 ID and 1:group 2 ID)
    elementUtilities.addEdge = function (
      source,
      target,
      edgeParams,
      id,
      visibility,
      groupID
    ) {

      console.log(source,
        target,
        edgeParams,
        id,
        visibility,
        groupID);
      if (typeof edgeParams != "object") {
        var sbgnclass = edgeParams;
      } else {
        var sbgnclass = edgeParams.class;
        var language = edgeParams.language;
      }

      var css = {};

      if (visibility) {
        css.visibility = visibility;
      }

      var data = {
        source: source,
        target: target,
        class: sbgnclass,
        language: language,
      };

      var defaults = elementUtilities.getDefaultProperties(sbgnclass);
      // extend the data with default properties of edge style
      Object.keys(defaults).forEach(function (prop) {
        data[prop] = defaults[prop];
      });

      if (id) {
        data.id = id;
      } else {
        data.id = elementUtilities.generateEdgeId();
      }

      if (elementUtilities.canHaveSBGNCardinality(sbgnclass)) {
        data.cardinality = 0;
      }
      var sourceNode = cy.getElementById(source); // The original source node
      var targetNode = cy.getElementById(target); // The original target node
      var sourceHasPorts = sourceNode.data("ports").length === 2;
      var targetHasPorts = targetNode.data("ports").length === 2;
      // The portsource and porttarget variables
      var portsource;
      var porttarget;


      /*
       * Get input/output port id's of a node with the assumption that the node has valid ports.
       */
      var getIOPortIds = function (node) {
        var nodeInputPortId, nodeOutputPortId;
        var nodePortsOrdering =
          sbgnvizInstance.elementUtilities.getPortsOrdering(node);
        var nodePorts = node.data("ports");
        if (nodePortsOrdering === "L-to-R" || nodePortsOrdering === "R-to-L") {
          var leftPortId =
            nodePorts[0].x < 0 ? nodePorts[0].id : nodePorts[1].id; // The x value of left port is supposed to be negative
          var rightPortId =
            nodePorts[0].x > 0 ? nodePorts[0].id : nodePorts[1].id; // The x value of right port is supposed to be positive
          /*
           * If the port ordering is left to right then the input port is the left port and the output port is the right port.
           * Else if it is right to left it is vice versa
           */
          nodeInputPortId =
            nodePortsOrdering === "L-to-R" ? leftPortId : rightPortId;
          nodeOutputPortId =
            nodePortsOrdering === "R-to-L" ? leftPortId : rightPortId;
        } else if (
          nodePortsOrdering === "T-to-B" ||
          nodePortsOrdering === "B-to-T"
        ) {
          var topPortId =
            nodePorts[0].y < 0 ? nodePorts[0].id : nodePorts[1].id; // The y value of top port is supposed to be negative
          var bottomPortId =
            nodePorts[0].y > 0 ? nodePorts[0].id : nodePorts[1].id; // The y value of bottom port is supposed to be positive
          /*
           * If the port ordering is top to bottom then the input port is the top port and the output port is the bottom port.
           * Else if it is right to left it is vice versa
           */
          nodeInputPortId =
            nodePortsOrdering === "T-to-B" ? topPortId : bottomPortId;
          nodeOutputPortId =
            nodePortsOrdering === "B-to-T" ? topPortId : bottomPortId;
        }

        // Return an object containing the IO ports of the node
        return {
          inputPortId: nodeInputPortId,
          outputPortId: nodeOutputPortId,
        };
      };
      // If at least one end of the edge has ports then we should determine the ports where the edge should be connected.
      if (sourceHasPorts || targetHasPorts) {
        var sourceNodeInputPortId,
          sourceNodeOutputPortId,
          targetNodeInputPortId,
          targetNodeOutputPortId;

        // If source node has ports set the variables dedicated for its IO ports
        if (sourceHasPorts) {
          var ioPorts = getIOPortIds(sourceNode);
          sourceNodeInputPortId = ioPorts.inputPortId;
          sourceNodeOutputPortId = ioPorts.outputPortId;
        }

        // && (elementUtilities.isModulationArcClass(sbgnclass) && )
        // If target node has ports set the variables dedicated for its IO ports
        if (targetHasPorts && !(elementUtilities.isModulationArcClass(sbgnclass) && elementUtilities.isProcessNode(targetNode.data('class')))) {
          var ioPorts = getIOPortIds(targetNode);
          targetNodeInputPortId = ioPorts.inputPortId;
          targetNodeOutputPortId = ioPorts.outputPortId;
        }

        if (
          sbgnclass === "consumption" ||
          sbgnclass === "translation consumption" ||
          sbgnclass === "transcription consumption"
        ) {
          // A consumption edge should be connected to the input port of the target node which is supposed to be a process (any kind of)
          portsource = sourceNodeOutputPortId;
          porttarget = targetNodeInputPortId;
        } else if (
          sbgnclass === "production" ||
          sbgnclass === "translation production" ||
          sbgnclass === "transcription production"||
          sbgnclass === "transport"
        ) {
          // A production edge should be connected to the output port of the source node which is supposed to be a process (any kind of)
          // A modulation edge may have a logical operator as source node in this case the edge should be connected to the output port of it
          // The below assignment satisfy all of these condition
          if (groupID == 0 || groupID == undefined) {
            // groupID 0 for reversible reactions group 0
            portsource = sourceNodeOutputPortId;
            porttarget = targetNodeInputPortId;
          } else {
            //if reaction is reversible and edge belongs to group 1
            portsource = sourceNodeInputPortId;
          }
        } else if (
          elementUtilities.isModulationArcClass(sbgnclass) ||
          elementUtilities.isAFArcClass(sbgnclass) ||
          elementUtilities.isSBMLArcClass(sbgnclass) ||
          elementUtilities.isLogicArc(sbgnclass)
        ) {
          portsource = sourceNodeOutputPortId;
          porttarget = targetNodeInputPortId;
        }
      }

      // The default portsource/porttarget are the source/target themselves. If they are not set use these defaults.
      // The portsource and porttarget are determined set them in data object.
      data.portsource = portsource || source;
      data.porttarget = porttarget || target;

      var eles = cy.add({
        group: "edges",
        data: data,
        css: css,
      });

      var newEdge = eles[eles.length - 1];

      return newEdge;
    };

    elementUtilities.addProcessWithConvenientEdges = function (
      _source,
      _target,
      nodeParams
    ) {
      // If source and target IDs are given get the elements by IDs
      var source =
        typeof _source === "string" ? cy.getElementById(_source) : _source;
      var target =
        typeof _target === "string" ? cy.getElementById(_target) : _target;

      // Process parent should be the closest common ancestor of the source and target nodes
      var processParent = cy
        .collection([source[0], target[0]])
        .commonAncestors()
        .first();

      // Process should be at the middle of the source and target nodes
      var x = (source.position("x") + target.position("x")) / 2;
      var y = (source.position("y") + target.position("y")) / 2;

      // Create the process with given/calculated variables
      var process = elementUtilities.addNode(
        x,
        y,
        nodeParams,
        undefined,
        processParent.id()
      );
      var xdiff = source.position("x") - target.position("x");
      var ydiff = source.position("y") - target.position("y");
      if (Math.abs(xdiff) >= Math.abs(ydiff)) {
        if (xdiff < 0) elementUtilities.setPortsOrdering(process, "L-to-R");
        else elementUtilities.setPortsOrdering(process, "R-to-L");
      } else {
        if (ydiff < 0) elementUtilities.setPortsOrdering(process, "T-to-B");
        else elementUtilities.setPortsOrdering(process, "B-to-T");
      }

      // Create the edges one is between the process and the source node (which should be a consumption),
      // the other one is between the process and the target node (which should be a production).
      // For more information please refer to SBGN-PD reference card.
      var edgeBtwSrc = elementUtilities.addEdge(source.id(), process.id(), {
        class: "consumption",
        language: nodeParams.language,
      });
      var edgeBtwTgt = elementUtilities.addEdge(process.id(), target.id(), {
        class: "production",
        language: nodeParams.language,
      });

      // Create a collection including the elements and to be returned
      var collection = cy.collection([
        process[0],
        edgeBtwSrc[0],
        edgeBtwTgt[0],
      ]);
      return collection;
    };

    /*
     * This method assumes that param.nodesToMakeCompound contains at least one node
     * and all of the nodes including in it have the same parent. It creates a compound fot the given nodes an having the given type.
     */
    elementUtilities.createCompoundForGivenNodes = function (
      nodesToMakeCompound,
      compoundType
    ) {
      var oldParentId = nodesToMakeCompound[0].data("parent");
      var language = nodesToMakeCompound[0].data("language");
      // if nodesToMakeCompound contain both PD and AF nodes, then set language of compound as Unknown
      for (var i = 1; i < nodesToMakeCompound.length; i++) {
        if (nodesToMakeCompound[i] != language) {
          language = "Unknown";
          break;
        }
      }
      // The parent of new compound will be the old parent of the nodes to make compound. x, y and id parameters are not set.
      var newCompound = elementUtilities.addNode(
        undefined,
        undefined,
        { class: compoundType, language: language },
        undefined,
        oldParentId
      );
      var newCompoundId = newCompound.id();
      var newEles = elementUtilities.changeParent(
        nodesToMakeCompound,
        newCompoundId
      );
      newEles = newEles.union(newCompound);
      return newEles;
    };

    elementUtilities.createTranslationReaction = function (
      mRnaName,
      proteinName,
      processPosition,
      edgeLength
    ) {
      const defaultProcessProperties =
        elementUtilities.getDefaultProperties("translation");
      const defaultSourceAndSinkProperties =
        elementUtilities.getDefaultProperties("empty set");
      const defaultNucleicAcidFeatureProperties =
        elementUtilities.getDefaultProperties("nucleic acid feature");
      const defaultMacromoleculeProperties =
        elementUtilities.getDefaultProperties("macromolecule");
      const macromoleculeWidth = defaultMacromoleculeProperties.width || 50;
      const sourceAndSinkWidth = defaultSourceAndSinkProperties.width || 50;
      const nucleicAcidFeatureHeight =
        defaultNucleicAcidFeatureProperties.height || 50;
      const processWidth = defaultProcessProperties.width || 50;
      const processHeight = defaultProcessProperties.height || 50;
      var processPosition =
        processPosition ||
        elementUtilities.convertToModelPosition({
          x: cy.width() / 2,
          y: cy.height() / 2,
        });
      var edgeLength = edgeLength || 60;

      cy.startBatch();
      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      var processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, "L-to-R");
      processNode.data("justAdded", true);

      const xPosOfSourceAndSinkNode =
        processPosition.x -
        edgeLength -
        processWidth / 2 -
        sourceAndSinkWidth / 2;
      const yPosOfSourceAndSinkNode = processPosition.y;
      var sourceAndSinkNode = elementUtilities.addNode(
        xPosOfSourceAndSinkNode,
        yPosOfSourceAndSinkNode,
        { class: "empty set", language: "PD" }
      );
      sourceAndSinkNode.data("justAdded", true);

      var consumptionEdge = elementUtilities.addEdge(
        sourceAndSinkNode.id(),
        processNode.id(),
        { class: "consumption", language: "PD" }
      );
      consumptionEdge.data("justAdded", true);

      const xPosOfmRnaNode = processPosition.x;
      const yPosOfmRnaNode =
        processPosition.y -
        edgeLength -
        processHeight / 2 -
        nucleicAcidFeatureHeight / 2;
      var mRnaNode = elementUtilities.addNode(xPosOfmRnaNode, yPosOfmRnaNode, {
        class: "nucleic acid feature",
        language: "PD",
      });
      mRnaNode.data("justAdded", true);
      mRnaNode.data("label", mRnaName);
      const infoboxObjectOfGene = {
        clazz: "unit of information",
        label: {
          text: "ct:mRNA",
        },
        bbox: {
          w: 45,
          h: 15,
        },
      };
      elementUtilities.addStateOrInfoBox(mRnaNode, infoboxObjectOfGene);

      var necessaryStimulationEdge = elementUtilities.addEdge(
        mRnaNode.id(),
        processNode.id(),
        { class: "necessary stimulation", language: "PD" }
      );
      necessaryStimulationEdge.data("justAdded", true);

      const xPosOfProteinNode =
        processPosition.x +
        edgeLength +
        processWidth / 2 +
        macromoleculeWidth / 2;
      const yPostOfProteinNode = processPosition.y;
      var proteinNode = elementUtilities.addNode(
        xPosOfProteinNode,
        yPostOfProteinNode,
        { class: "macromolecule", language: "PD" }
      );
      proteinNode.data("justAdded", true);
      proteinNode.data("label", proteinName);

      var productionEdge = elementUtilities.addEdge(
        processNode.id(),
        proteinNode.id(),
        { class: "production", language: "PD" }
      );
      productionEdge.data("justAdded", true);

      cy.endBatch();

      //filter the just added elememts to return them and remove just added mark
      var eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles; // Return the just added elements
    };

    elementUtilities.createTranscriptionReaction = function (
      geneName,
      mRnaName,
      processPosition,
      edgeLength
    ) {
      const defaultProcessProperties =
        elementUtilities.getDefaultProperties("transcription");
      const defaultSourceAndSinkProperties =
        elementUtilities.getDefaultProperties("empty set");
      const defaultNucleicAcidFeatureProperties =
        elementUtilities.getDefaultProperties("nucleic acid feature");
      const sourceAndSinkWidth = defaultSourceAndSinkProperties.width || 50;
      const nucleicAcidFeatureHeight =
        defaultNucleicAcidFeatureProperties.height || 50;
      const nucleicAcidFeatureWidth =
        defaultNucleicAcidFeatureProperties.width || 50;
      const processWidth = defaultProcessProperties.width || 50;
      const processHeight = defaultProcessProperties.height || 50;
      var processPosition =
        processPosition ||
        elementUtilities.convertToModelPosition({
          x: cy.width() / 2,
          y: cy.height() / 2,
        });
      var edgeLength = edgeLength || 60;

      cy.startBatch();
      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      var processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, "L-to-R");
      processNode.data("justAdded", true);

      const xPosOfSourceAndSinkNode =
        processPosition.x -
        edgeLength -
        processWidth / 2 -
        sourceAndSinkWidth / 2;
      const yPosOfSourceAndSinkNode = processPosition.y;
      var sourceAndSinkNode = elementUtilities.addNode(
        xPosOfSourceAndSinkNode,
        yPosOfSourceAndSinkNode,
        { class: "empty set", language: "PD" }
      );
      sourceAndSinkNode.data("justAdded", true);

      var consumptionEdge = elementUtilities.addEdge(
        sourceAndSinkNode.id(),
        processNode.id(),
        { class: "consumption", language: "PD" }
      );
      consumptionEdge.data("justAdded", true);

      const xPosOfGeneNode = processPosition.x;
      const yPosOfGeneNode =
        processPosition.y -
        edgeLength -
        processHeight / 2 -
        nucleicAcidFeatureHeight / 2;
      var geneNode = elementUtilities.addNode(xPosOfGeneNode, yPosOfGeneNode, {
        class: "nucleic acid feature",
        language: "PD",
      });
      geneNode.data("justAdded", true);
      geneNode.data("label", geneName);
      const infoboxObjectOfGene = {
        clazz: "unit of information",
        label: {
          text: "ct:gene",
        },
        bbox: {
          w: 36,
          h: 15,
        },
      };
      elementUtilities.addStateOrInfoBox(geneNode, infoboxObjectOfGene);

      var necessaryStimulationEdge = elementUtilities.addEdge(
        geneNode.id(),
        processNode.id(),
        { class: "necessary stimulation", language: "PD" }
      );
      necessaryStimulationEdge.data("justAdded", true);

      const xPosOfmRnaNode =
        processPosition.x +
        edgeLength +
        processWidth / 2 +
        nucleicAcidFeatureWidth / 2;
      const yPostOfmRnaNode = processPosition.y;
      var mRnaNode = elementUtilities.addNode(xPosOfmRnaNode, yPostOfmRnaNode, {
        class: "nucleic acid feature",
        language: "PD",
      });
      mRnaNode.data("justAdded", true);
      mRnaNode.data("label", mRnaName);
      const infoboxObjectOfmRna = {
        clazz: "unit of information",
        label: {
          text: "ct:mRNA",
        },
        bbox: {
          w: 45,
          h: 15,
        },
      };
      elementUtilities.addStateOrInfoBox(mRnaNode, infoboxObjectOfmRna);

      var productionEdge = elementUtilities.addEdge(
        processNode.id(),
        mRnaNode.id(),
        { class: "production", language: "PD" }
      );
      productionEdge.data("justAdded", true);

      cy.endBatch();

      //filter the just added elememts to return them and remove just added mark
      var eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles; // Return the just added elements
    };

    elementUtilities.rotate90 = function (point, center) {
      const relativeX = center.x - point.x;
      const relativeY = center.y - point.y;

      const relativeRotatedX = relativeY;
      const relativeRotatedY = -1 * relativeX;

      const resultX = relativeRotatedX + center.x;
      const resultY = relativeRotatedY + center.y;

      return {
        x: resultX,
        y: resultY,
      };
    };

    elementUtilities.createTranslation = function (
      regulatorLabel,
      outputLabel,
      orientation
    ) {
      const defaultSourceAndSinkProperties =
        elementUtilities.getDefaultProperties("empty set");
      const defaultNucleicAcidFeatureProperties =
        elementUtilities.getDefaultProperties("nucleic acid feature");
      const defaultMacromoleculePropeties =
        elementUtilities.getDefaultProperties("macromolecule");
      const defaultProcessProperties =
        elementUtilities.getDefaultProperties("process");
      const sourceAndSinkWidth = defaultSourceAndSinkProperties.width || 50;
      const nucleicAcidFeatureWidth =
        defaultNucleicAcidFeatureProperties.width || 50;
      const nucleicAcidFeatureHeight =
        defaultNucleicAcidFeatureProperties.height || 50;
      const macromoleculeWidth = defaultMacromoleculePropeties.width || 50;
      const processWidth = defaultProcessProperties.width || 50;
      const processHeight = defaultProcessProperties.height || 50;
      const processPosition = elementUtilities.convertToModelPosition({
        x: cy.width() / 2,
        y: cy.height() / 2,
      });
      const edgeLength = 30;
      const vertical = orientation === "vertical";
      const processPortsOrdering = vertical ? "T-to-B" : "L-to-R";
      const minInfoboxDimension = 15;
      const widthPerChar = 6;
      const regulatorInfoboxLabel = "ct:mRNA";

      cy.startBatch();

      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }
      const processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, processPortsOrdering);
      processNode.data("justAdded", true);

      let xPosOfInput =
        processPosition.x -
        edgeLength -
        processWidth / 2 -
        sourceAndSinkWidth / 2;
      let xPosOfOutput =
        processPosition.x +
        edgeLength +
        processWidth / 2 +
        macromoleculeWidth / 2;
      let yPosOfInput = processPosition.y;
      let yPosOfOutput = processPosition.y;

      let nodePosition = {
        x: xPosOfInput,
        y: yPosOfInput,
      };
      if (vertical) {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      const inputNode = elementUtilities.addNode(
        nodePosition.x,
        nodePosition.y,
        { class: "empty set", language: "PD" }
      );
      inputNode.data("justAdded", true);
      inputNode.data("label", label);

      const inputEdge = elementUtilities.addEdge(
        inputNode.id(),
        processNode.id(),
        { class: "consumption", language: "PD" }
      );
      inputEdge.data("justAdded", true);

      nodePosition = {
        x: xPosOfOutput,
        y: yPosOfOutput,
      };

      if (vertical) {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      const outputNode = elementUtilities.addNode(
        nodePosition.x,
        nodePosition.y,
        { class: "macromolecule", language: "PD" }
      );
      outputNode.data("justAdded", true);
      outputNode.data("label", outputLabel);

      const outputEdge = elementUtilities.addEdge(
        processNode.id(),
        outputNode.id(),
        { class: "production", language: "PD" }
      );
      outputEdge.data("justAdded", true);

      let xPosOfRegulator = processPosition.x;
      const dimension = vertical
        ? nucleicAcidFeatureWidth
        : nucleicAcidFeatureHeight;
      let yPosOfRegulator =
        processPosition.y - (processHeight / 2 + dimension / 2 + edgeLength);

      nodePosition = {
        x: xPosOfRegulator,
        y: yPosOfRegulator,
      };
      if (vertical) {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      const regulatorNode = elementUtilities.addNode(
        nodePosition.x,
        nodePosition.y,
        { class: "nucleic acid feature", language: "PD" }
      );
      regulatorNode.data("justAdded", true);
      regulatorNode.data("label", regulatorLabel);
      infoboxObject = {
        clazz: "unit of information",
        label: {
          text: regulatorInfoboxLabel,
        },
        bbox: {
          w: Math.max(
            regulatorInfoboxLabel.length * widthPerChar,
            minInfoboxDimension
          ),
          h: minInfoboxDimension,
        },
      };
      elementUtilities.addStateOrInfoBox(regulatorNode, infoboxObject);

      const regulatorEdge = elementUtilities.addEdge(
        regulatorNode.id(),
        processNode.id(),
        { class: "necessary stimulation", language: "PD" }
      );
      regulatorEdge.data("justAdded", true);

      cy.endBatch();

      const eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles;
    };

    elementUtilities.createTranscription = function (label, orientation) {
      const defaultSourceAndSinkProperties =
        elementUtilities.getDefaultProperties("empty set");
      const defaultNucleicAcidFeatureProperties =
        elementUtilities.getDefaultProperties("nucleic acid feature");
      const defaultProcessProperties =
        elementUtilities.getDefaultProperties("process");
      const sourceAndSinkWidth = defaultSourceAndSinkProperties.width || 50;
      const nucleicAcidFeatureWidth =
        defaultNucleicAcidFeatureProperties.width || 50;
      const nucleicAcidFeatureHeight =
        defaultNucleicAcidFeatureProperties.height || 50;
      const processWidth = defaultProcessProperties.width || 50;
      const processHeight = defaultProcessProperties.height || 50;
      const processPosition = elementUtilities.convertToModelPosition({
        x: cy.width() / 2,
        y: cy.height() / 2,
      });
      const edgeLength = 30;
      const vertical = orientation === "vertical";
      const processPortsOrdering = vertical ? "T-to-B" : "L-to-R";
      const minInfoboxDimension = 15;
      const widthPerChar = 6;
      const outputInfoboxLabel = "ct:mRNA";
      const regulatorInfoboxLabel = "ct:gene";

      cy.startBatch();

      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      const processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, processPortsOrdering);
      processNode.data("justAdded", true);

      let xPosOfInput =
        processPosition.x -
        edgeLength -
        processWidth / 2 -
        sourceAndSinkWidth / 2;
      let xPosOfOutput =
        processPosition.x +
        edgeLength +
        processWidth / 2 +
        nucleicAcidFeatureWidth / 2;
      let yPosOfInput = processPosition.y;
      let yPosOfOutput = processPosition.y;

      let nodePosition = {
        x: xPosOfInput,
        y: yPosOfInput,
      };
      if (vertical) {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      const inputNode = elementUtilities.addNode(
        nodePosition.x,
        nodePosition.y,
        { class: "empty set", language: "PD" }
      );
      inputNode.data("justAdded", true);

      const inputEdge = elementUtilities.addEdge(
        inputNode.id(),
        processNode.id(),
        { class: "consumption", language: "PD" }
      );
      inputEdge.data("justAdded", true);

      nodePosition = {
        x: xPosOfOutput,
        y: yPosOfOutput,
      };

      if (vertical) {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      const outputNode = elementUtilities.addNode(
        nodePosition.x,
        nodePosition.y,
        { class: "nucleic acid feature", language: "PD" }
      );
      outputNode.data("justAdded", true);
      outputNode.data("label", label);
      infoboxObject = {
        clazz: "unit of information",
        label: {
          text: outputInfoboxLabel,
        },
        bbox: {
          w: Math.max(
            outputInfoboxLabel.length * widthPerChar,
            minInfoboxDimension
          ),
          h: minInfoboxDimension,
        },
      };
      elementUtilities.addStateOrInfoBox(outputNode, infoboxObject);

      const outputEdge = elementUtilities.addEdge(
        processNode.id(),
        outputNode.id(),
        { class: "production", language: "PD" }
      );
      outputEdge.data("justAdded", true);

      let xPosOfRegulator = processPosition.x;
      const dimension = vertical
        ? nucleicAcidFeatureWidth
        : nucleicAcidFeatureHeight;
      let yPosOfRegulator =
        processPosition.y - (processHeight / 2 + dimension / 2 + edgeLength);

      nodePosition = {
        x: xPosOfRegulator,
        y: yPosOfRegulator,
      };
      if (vertical) {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      const regulatorNode = elementUtilities.addNode(
        nodePosition.x,
        nodePosition.y,
        { class: "nucleic acid feature", language: "PD" }
      );
      regulatorNode.data("justAdded", true);
      regulatorNode.data("label", label);
      infoboxObject = {
        clazz: "unit of information",
        label: {
          text: regulatorInfoboxLabel,
        },
        bbox: {
          w: Math.max(
            regulatorInfoboxLabel.length * widthPerChar,
            minInfoboxDimension
          ),
          h: minInfoboxDimension,
        },
      };
      elementUtilities.addStateOrInfoBox(regulatorNode, infoboxObject);

      const regulatorEdge = elementUtilities.addEdge(
        regulatorNode.id(),
        processNode.id(),
        { class: "necessary stimulation", language: "PD" }
      );
      regulatorEdge.data("justAdded", true);

      cy.endBatch();

      const eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles;
    };

    elementUtilities.createDegradation = function (macromolecule, orientation) {
      const macromoleculeName = macromolecule.name;
      const defaultMacromoleculeProperties =
        elementUtilities.getDefaultProperties("macromolecule");
      const macromoleculeWidth = defaultMacromoleculeProperties.width || 50;
      const defaultProcessProperties =
        elementUtilities.getDefaultProperties("catalytic");
      const processWidth = defaultProcessProperties.width || 50;
      const processPosition = elementUtilities.convertToModelPosition({
        x: cy.width() / 2,
        y: cy.height() / 2,
      });
      const edgeLength = 30;
      const vertical = orientation === "vertical";
      const processPortsOrdering = vertical ? "T-to-B" : "L-to-R";

      cy.startBatch();

      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      const processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, processPortsOrdering);
      processNode.data("justAdded", true);

      let xPosOfInput =
        processPosition.x -
        edgeLength -
        processWidth / 2 -
        macromoleculeWidth / 2;
      let xPosOfOutput =
        processPosition.x +
        edgeLength +
        processWidth / 2 +
        macromoleculeWidth / 2;
      let yPosOfInput = processPosition.y;
      let yPosOfOutput = processPosition.y;

      let nodePosition = {
        x: xPosOfInput,
        y: yPosOfInput,
      };
      if (vertical) {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      let inputNode = elementUtilities.addNode(nodePosition.x, nodePosition.y, {
        class: "macromolecule",
        language: "PD",
      });
      inputNode.data("justAdded", true);
      inputNode.data("label", macromoleculeName);

      let inputEdge = elementUtilities.addEdge(
        inputNode.id(),
        processNode.id(),
        { class: "consumption", language: "PD" }
      );
      inputEdge.data("justAdded", true);

      nodePosition = {
        x: xPosOfOutput,
        y: yPosOfOutput,
      };

      if (vertical) {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      let outputNode = elementUtilities.addNode(
        nodePosition.x,
        nodePosition.y,
        { class: "empty set", language: "PD" }
      );
      outputNode.data("justAdded", true);

      let outputEdge = elementUtilities.addEdge(
        processNode.id(),
        outputNode.id(),
        { class: "production", language: "PD" }
      );
      outputEdge.data("justAdded", true);

      cy.endBatch();

      const eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles;
    };

    elementUtilities.createComplexProteinFormation = function (
      proteinLabels,
      complexLabel,
      regulator,
      orientation,
      reverse
    ) {
      const hasRegulator = regulator.name !== undefined;
      const defaultMacromoleculeProperties =
        elementUtilities.getDefaultProperties("macromolecule");
      const defaultRegulatorProperties = hasRegulator
        ? elementUtilities.getDefaultProperties(regulator.type)
        : {};
      const defaultProcessProperties =
        elementUtilities.getDefaultProperties("catalytic");
      const processWidth = defaultProcessProperties.width || 50;
      const macromoleculeWidth = defaultMacromoleculeProperties.width || 50;
      const macromoleculeHeight = defaultMacromoleculeProperties.height || 50;
      const processHeight = defaultProcessProperties.height || 50;
      const regulatorHeight = defaultRegulatorProperties.height || 50;
      const processPosition = elementUtilities.convertToModelPosition({
        x: cy.width() / 2,
        y: cy.height() / 2,
      });
      const edgeLength = 30;
      const processPortsOrdering =
        orientation === "vertical" ? "T-to-B" : "L-to-R";
      const minInfoboxDimension = 20;
      const widthPerChar = 6;
      const tilingPaddingVertical = 15;
      const tilingPaddingHorizontal = 15;
      const multimerOffset = 6;

      cy.startBatch();

      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      const processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, processPortsOrdering);
      processNode.data("justAdded", true);

      const offsetX = processWidth / 2 + edgeLength + macromoleculeWidth / 2;
      let xPosOfProtein = reverse
        ? processPosition.x + offsetX
        : processPosition.x - offsetX;

      const proteinCount = proteinLabels.length;

      const macromoleculeDimension =
        orientation === "vertical" ? macromoleculeWidth : macromoleculeHeight;
      const stepOffset = macromoleculeDimension + tilingPaddingVertical;
      const offsetY =
        ((proteinCount - 1) / 2) *
        (macromoleculeDimension + tilingPaddingVertical);
      const horizontalOffsetX =
        ((proteinCount - 1) / 2) *
        (macromoleculeDimension + tilingPaddingHorizontal);

      let yPosOfProtein = processPosition.y - offsetY;

      proteinLabels.forEach(function (label) {
        let nodePosition = {
          x: xPosOfProtein,
          y: yPosOfProtein,
        };
        if (orientation === "vertical") {
          nodePosition = elementUtilities.rotate90(
            nodePosition,
            processPosition
          );
        }

        const node = elementUtilities.addNode(nodePosition.x, nodePosition.y, {
          class: "macromolecule",
          language: "PD",
        });
        node.data("label", label);
        node.data("justAdded", true);
        yPosOfProtein += stepOffset;

        const source = reverse ? processNode.id() : node.id();
        const target = reverse ? node.id() : processNode.id();
        const edgeClass = reverse ? "production" : "consumption";
        const edge = elementUtilities.addEdge(source, target, {
          class: edgeClass,
          language: "PD",
        });
        edge.data("justAdded", true);
      });

      let complexPos = {
        x: processPosition.x + (reverse ? -1 : 1) * offsetX,
        y: processPosition.y,
      };

      if (orientation === "vertical") {
        complexPos = elementUtilities.rotate90(complexPos, processPosition);
      }

      const complex = elementUtilities.addNode(complexPos.x, complexPos.y, {
        class: "complex",
        language: "PD",
      });
      complex.data("label", complexLabel);
      complex.data("justAdded", true);

      const source = reverse ? complex.id() : processNode.id();
      const target = reverse ? processNode.id() : complex.id();
      const edgeClass = reverse ? "consumption" : "production";
      const complexEdge = elementUtilities.addEdge(source, target, {
        class: edgeClass,
        language: "PD",
      });
      complexEdge.data("justAdded", true);

      if (orientation === "vertical") {
        xPosOfProtein = complex.position("x") - horizontalOffsetX;
        yPosOfProtein = complex.position("y");
      } else {
        xPosOfProtein = complex.position("x");
        yPosOfProtein = complex.position("y") - offsetY;
      }

      proteinLabels.forEach(function (label) {
        let nodePosition = {
          x: xPosOfProtein,
          y: yPosOfProtein,
        };

        const node = elementUtilities.addNode(
          nodePosition.x,
          nodePosition.y,
          { class: "macromolecule", language: "PD" },
          undefined,
          complex.id()
        );
        node.data("label", label);
        node.data("justAdded", true);

        if (orientation === "vertical") {
          xPosOfProtein += stepOffset;
        } else {
          yPosOfProtein += stepOffset;
        }
      });

      if (hasRegulator) {
        const regulatorName = regulator.name;
        const regulatorType = regulator.type;
        const regulatorEdgeType = regulator.edgeType;
        const regulatorMultimer = regulator.multimer;

        let xPosOfRegulator = processPosition.x;
        if (regulatorMultimer.enabled && orientation === "horizontal") {
          xPosOfRegulator -= multimerOffset;
        }
        let yPosOfRegulator =
          processPosition.y -
          (processHeight / 2 + regulatorHeight / 2 + edgeLength);

        nodePosition = {
          x: xPosOfRegulator,
          y: yPosOfRegulator,
        };
        if (orientation === "vertical") {
          nodePosition = elementUtilities.rotate90(
            nodePosition,
            processPosition
          );
        }

        let regulatorNode = elementUtilities.addNode(
          nodePosition.x,
          nodePosition.y,
          { class: regulatorType, language: "PD" }
        );
        regulatorNode.data("justAdded", true);
        regulatorNode.data("label", regulatorName);

        if (regulatorMultimer.enabled) {
          elementUtilities.setMultimerStatus(regulatorNode, true);

          const cardinality = regulatorMultimer.cardinality;
          if (cardinality != "") {
            const infoboxLabel = "N:" + cardinality;
            infoboxObject = {
              clazz: "unit of information",
              label: {
                text: infoboxLabel,
              },
              bbox: {
                w: infoboxLabel.length * widthPerChar,
                h: minInfoboxDimension,
              },
            };
            elementUtilities.addStateOrInfoBox(regulatorNode, infoboxObject);
          }
        }

        let regulatorEdge = elementUtilities.addEdge(
          regulatorNode.id(),
          processNode.id(),
          { class: regulatorEdgeType, language: "PD" }
        );
        regulatorEdge.data("justAdded", true);
      }

      cy.endBatch();

      const eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles;
    };

    elementUtilities.createMultimerization = function (
      macromolecule,
      regulator,
      regulatorMultimer,
      orientation
    ) {
      const hasRegulator = regulator.name !== undefined;
      const macromoleculeName = macromolecule.name;
      const macromoleculeMultimerCardinality = macromolecule.cardinality;
      const defaultMacromoleculeProperties =
        elementUtilities.getDefaultProperties("macromolecule");
      const defaultRegulatorProperties = hasRegulator
        ? elementUtilities.getDefaultProperties(regulator.type)
        : {};
      const defaultProcessProperties =
        elementUtilities.getDefaultProperties("catalytic");
      const processWidth = defaultProcessProperties.width || 50;
      const macromoleculeWidth = defaultMacromoleculeProperties.width || 50;
      const macromoleculeHeight = defaultMacromoleculeProperties.height || 50;
      const processHeight = defaultProcessProperties.height || 50;
      const regulatorHeight = defaultRegulatorProperties.height || 50;
      const processPosition = elementUtilities.convertToModelPosition({
        x: cy.width() / 2,
        y: cy.height() / 2,
      });
      const edgeLength = 30;
      const processPortsOrdering =
        orientation === "vertical" ? "T-to-B" : "L-to-R";
      const minInfoboxDimension = 20;
      const widthPerChar = 6;
      const multimerOffset = 6;

      cy.startBatch();

      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      let xPosOfInput =
        processPosition.x -
        edgeLength -
        processWidth / 2 -
        macromoleculeWidth / 2;
      let xPosOfOutput =
        processPosition.x +
        edgeLength +
        processWidth / 2 +
        macromoleculeWidth / 2;
      let yPosOfInput = processPosition.y;
      let yPosOfOutput = processPosition.y;

      let processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, processPortsOrdering);
      processNode.data("justAdded", true);

      let nodePosition = {
        x: xPosOfInput,
        y: yPosOfInput,
      };
      if (orientation === "vertical") {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      let inputNode = elementUtilities.addNode(nodePosition.x, nodePosition.y, {
        class: "macromolecule",
        language: "PD",
      });
      inputNode.data("justAdded", true);
      inputNode.data("label", macromoleculeName);

      let inputEdge = elementUtilities.addEdge(
        inputNode.id(),
        processNode.id(),
        { class: "consumption", language: "PD" }
      );
      inputEdge.data("justAdded", true);

      let cardinality = macromoleculeMultimerCardinality;
      if (cardinality !== "") {
        inputEdge.data("cardinality", cardinality);
      }

      nodePosition = {
        x: xPosOfOutput,
        y: yPosOfOutput,
      };

      if (orientation === "vertical") {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      let outputNode = elementUtilities.addNode(
        nodePosition.x,
        nodePosition.y,
        { class: "macromolecule", language: "PD" }
      );
      outputNode.data("justAdded", true);
      outputNode.data("label", macromoleculeName);
      elementUtilities.setMultimerStatus(outputNode, true);

      if (cardinality !== "") {
        const infoboxLabel = "N:" + cardinality;
        infoboxObject = {
          clazz: "unit of information",
          label: {
            text: infoboxLabel,
          },
          bbox: {
            w: infoboxLabel.length * widthPerChar,
            h: minInfoboxDimension,
          },
        };
        elementUtilities.addStateOrInfoBox(outputNode, infoboxObject);
      }

      let outputEdge = elementUtilities.addEdge(
        processNode.id(),
        outputNode.id(),
        { class: "production", language: "PD" }
      );
      outputEdge.data("justAdded", true);

      if (hasRegulator) {
        const regulatorName = regulator.name;
        const regulatorType = regulator.type;
        const regulatorEdgeType = regulator.edgeType;

        let xPosOfRegulator = processPosition.x;
        if (regulatorMultimer.enabled && orientation === "horizontal") {
          xPosOfRegulator -= multimerOffset;
        }
        let yPosOfRegulator =
          processPosition.y -
          (processHeight / 2 + regulatorHeight / 2 + edgeLength);

        nodePosition = {
          x: xPosOfRegulator,
          y: yPosOfRegulator,
        };
        if (orientation === "vertical") {
          nodePosition = elementUtilities.rotate90(
            nodePosition,
            processPosition
          );
        }

        let regulatorNode = elementUtilities.addNode(
          nodePosition.x,
          nodePosition.y,
          { class: regulatorType, language: "PD" }
        );
        regulatorNode.data("justAdded", true);
        regulatorNode.data("label", regulatorName);

        if (regulatorMultimer.enabled) {
          elementUtilities.setMultimerStatus(regulatorNode, true);

          const cardinality = regulatorMultimer.cardinality;
          if (cardinality != "") {
            const infoboxLabel = "N:" + cardinality;
            infoboxObject = {
              clazz: "unit of information",
              label: {
                text: infoboxLabel,
              },
              bbox: {
                w: infoboxLabel.length * widthPerChar,
                h: minInfoboxDimension,
              },
            };
            elementUtilities.addStateOrInfoBox(regulatorNode, infoboxObject);
          }
        }

        let regulatorEdge = elementUtilities.addEdge(
          regulatorNode.id(),
          processNode.id(),
          { class: regulatorEdgeType, language: "PD" }
        );
        regulatorEdge.data("justAdded", true);
      }

      cy.endBatch();

      const eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles;
    };

    elementUtilities.createConversion = function (
      macromolecule,
      regulator,
      regulatorMultimer,
      orientation,
      inputInfoboxLabels,
      outputInfoboxLabels
    ) {
      const hasRegulator = regulator.name !== undefined;
      const macromoleculeName = macromolecule.name;
      const macromoleculeIsMultimer = macromolecule.multimer.enabled;
      const macromoleculeMultimerCardinality =
        macromolecule.multimer.cardinality;
      const defaultMacromoleculeProperties =
        elementUtilities.getDefaultProperties("macromolecule");
      const defaultRegulatorProperties = hasRegulator
        ? elementUtilities.getDefaultProperties(regulator.type)
        : {};
      const defaultProcessProperties =
        elementUtilities.getDefaultProperties("catalytic");
      const processWidth = defaultProcessProperties.width || 50;
      const macromoleculeWidth = defaultMacromoleculeProperties.width || 50;
      const macromoleculeHeight = defaultMacromoleculeProperties.height || 50;
      const processHeight = defaultProcessProperties.height || 50;
      const regulatorHeight = defaultRegulatorProperties.height || 50;
      const processPosition = elementUtilities.convertToModelPosition({
        x: cy.width() / 2,
        y: cy.height() / 2,
      });
      const edgeLength = 30;
      const processPortsOrdering =
        orientation === "vertical" ? "T-to-B" : "L-to-R";
      const minInfoboxDimension = 20;
      const widthPerChar = 6;
      const multimerOffset = 6;

      cy.startBatch();

      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      let xPosOfInput =
        processPosition.x -
        edgeLength -
        processWidth / 2 -
        macromoleculeWidth / 2;
      let xPosOfOutput =
        processPosition.x +
        edgeLength +
        processWidth / 2 +
        macromoleculeWidth / 2;
      let yPosOfInput = processPosition.y;
      let yPosOfOutput = processPosition.y;

      let processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, processPortsOrdering);
      processNode.data("justAdded", true);

      let nodePosition = {
        x: xPosOfInput,
        y: yPosOfInput,
      };
      if (orientation === "vertical") {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      let inputNode = elementUtilities.addNode(nodePosition.x, nodePosition.y, {
        class: "macromolecule",
        language: "PD",
      });
      inputNode.data("justAdded", true);
      inputNode.data("label", macromoleculeName);
      if (macromoleculeIsMultimer) {
        elementUtilities.setMultimerStatus(inputNode, true);

        const cardinality = macromoleculeMultimerCardinality;
        if (cardinality != "") {
          const infoboxLabel = "N:" + cardinality;
          infoboxObject = {
            clazz: "unit of information",
            label: {
              text: infoboxLabel,
            },
            bbox: {
              w: infoboxLabel.length * widthPerChar,
              h: minInfoboxDimension,
            },
          };
          elementUtilities.addStateOrInfoBox(inputNode, infoboxObject);
        }
      }

      inputInfoboxLabels.forEach(function (label) {
        const inputInfoboxWidth =
          label.length > 0
            ? Math.max(widthPerChar * label.length, minInfoboxDimension)
            : minInfoboxDimension;
        let infoboxObject = {
          clazz: "unit of information",
          label: {
            text: label,
          },
          bbox: {
            w: inputInfoboxWidth,
            h: minInfoboxDimension,
          },
          style: {
            "shape-name": "ellipse",
          },
        };
        elementUtilities.addStateOrInfoBox(inputNode, infoboxObject);
      });

      let inputEdge = elementUtilities.addEdge(
        inputNode.id(),
        processNode.id(),
        { class: "consumption", language: "PD" }
      );
      inputEdge.data("justAdded", true);

      nodePosition = {
        x: xPosOfOutput,
        y: yPosOfOutput,
      };

      if (orientation === "vertical") {
        nodePosition = elementUtilities.rotate90(nodePosition, processPosition);
      }

      let outputNode = elementUtilities.addNode(
        nodePosition.x,
        nodePosition.y,
        { class: "macromolecule", language: "PD" }
      );
      outputNode.data("justAdded", true);
      outputNode.data("label", macromoleculeName);
      if (macromoleculeIsMultimer) {
        elementUtilities.setMultimerStatus(outputNode, true);

        const cardinality = macromoleculeMultimerCardinality;
        if (cardinality != "") {
          const infoboxLabel = "N:" + cardinality;
          infoboxObject = {
            clazz: "unit of information",
            label: {
              text: infoboxLabel,
            },
            bbox: {
              w: infoboxLabel.length * widthPerChar,
              h: minInfoboxDimension,
            },
          };
          elementUtilities.addStateOrInfoBox(outputNode, infoboxObject);
        }
      }

      outputInfoboxLabels.forEach(function (label) {
        const outputInfoboxWidth =
          label.length > 0
            ? Math.max(widthPerChar * label.length, minInfoboxDimension)
            : minInfoboxDimension;
        infoboxObject = {
          clazz: "unit of information",
          label: {
            text: label,
          },
          bbox: {
            w: outputInfoboxWidth,
            h: minInfoboxDimension,
          },
          style: {
            "shape-name": "ellipse",
          },
        };
        elementUtilities.addStateOrInfoBox(outputNode, infoboxObject);
      });

      [inputNode, outputNode].forEach(function (node) {
        const width = elementUtilities.calculateMinWidth(node);

        elementUtilities.resizeNodes(
          node,
          width,
          macromoleculeHeight,
          false,
          true
        );
      });

      if (orientation === "horizontal") {
        let newInputXPos =
          processPosition.x -
          edgeLength -
          processWidth / 2 -
          inputNode.data("bbox").w / 2;
        inputNode.position("x", newInputXPos);

        let newOutputXPos =
          processPosition.x +
          edgeLength +
          processWidth / 2 +
          outputNode.data("bbox").w / 2;
        outputNode.position("x", newOutputXPos);
      } else {
        let newInputYPos =
          processPosition.y -
          edgeLength -
          processWidth / 2 -
          inputNode.data("bbox").h / 2;
        inputNode.position("y", newInputYPos);

        let newOutputYPos =
          processPosition.y +
          edgeLength +
          processWidth / 2 +
          outputNode.data("bbox").h / 2;
        outputNode.position("y", newOutputYPos);
      }

      let outputEdge = elementUtilities.addEdge(
        processNode.id(),
        outputNode.id(),
        { class: "production", language: "PD" }
      );
      outputEdge.data("justAdded", true);

      if (hasRegulator) {
        const regulatorName = regulator.name;
        const regulatorType = regulator.type;
        let xPosOfRegulator = processPosition.x;
        if (regulatorMultimer.enabled && orientation === "horizontal") {
          xPosOfRegulator -= multimerOffset;
        }
        let yPosOfRegulator =
          processPosition.y -
          (processHeight / 2 + regulatorHeight / 2 + edgeLength);

        nodePosition = {
          x: xPosOfRegulator,
          y: yPosOfRegulator,
        };
        if (orientation === "vertical") {
          nodePosition = elementUtilities.rotate90(
            nodePosition,
            processPosition
          );
        }

        let regulatorNode = elementUtilities.addNode(
          nodePosition.x,
          nodePosition.y,
          { class: regulatorType, language: "PD" }
        );
        regulatorNode.data("justAdded", true);
        regulatorNode.data("label", regulatorName);

        if (regulatorMultimer.enabled) {
          elementUtilities.setMultimerStatus(regulatorNode, true);

          const cardinality = regulatorMultimer.cardinality;
          if (cardinality != "") {
            const infoboxLabel = "N:" + cardinality;
            infoboxObject = {
              clazz: "unit of information",
              label: {
                text: infoboxLabel,
              },
              bbox: {
                w: infoboxLabel.length * widthPerChar,
                h: minInfoboxDimension,
              },
            };
            elementUtilities.addStateOrInfoBox(regulatorNode, infoboxObject);
          }
        }

        let regulatorEdge = elementUtilities.addEdge(
          regulatorNode.id(),
          processNode.id(),
          { class: "catalysis", language: "PD" }
        );
        regulatorEdge.data("justAdded", true);
      }

      cy.endBatch();

      const eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles;
    };

    elementUtilities.createMetabolicReaction = function (
      inputs,
      outputs,
      reversible,
      regulator,
      regulatorMultimer,
      orientation
    ) {
      let rotate90 = function (point, center) {
        const relativeX = center.x - point.x;
        const relativeY = center.y - point.y;

        const relativeRotatedX = relativeY;
        const relativeRotatedY = -1 * relativeX;

        const resultX = relativeRotatedX + center.x;
        const resultY = relativeRotatedY + center.y;

        return {
          x: resultX,
          y: resultY,
        };
      };
      const hasRegulator = regulator.name !== undefined;
      const defaultSimpleChemicalProperties =
        elementUtilities.getDefaultProperties("simple chemical");
      const defaultRegulatorProperties = hasRegulator
        ? elementUtilities.getDefaultProperties(regulator.type)
        : {};
      const defaultProcessProperties =
        elementUtilities.getDefaultProperties("catalytic");
      const processWidth = defaultProcessProperties.width || 50;
      const processHeight = defaultProcessProperties.height || 50;
      const simpleChemicalHeight = defaultSimpleChemicalProperties.height || 35;
      const simpleChemicalWidth = defaultSimpleChemicalProperties.width || 35;
      const regulatorHeight = defaultRegulatorProperties.height || 50;
      const processPosition = elementUtilities.convertToModelPosition({
        x: cy.width() / 2,
        y: cy.height() / 2,
      });
      const tilingPaddingVertical = 15;
      const edgeLength = 30;
      const processLeftSideEdgeType = reversible ? "production" : "consumption";
      const processRightSideEdgeType = "production";
      const processPortsOrdering =
        orientation === "vertical" ? "T-to-B" : "L-to-R";
      const multimerOffset = 6;

      cy.startBatch();
      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      let xPosOfInput =
        processPosition.x -
        edgeLength -
        processWidth / 2 -
        simpleChemicalWidth / 2;
      let xPosOfOutput =
        processPosition.x +
        edgeLength +
        processWidth / 2 +
        simpleChemicalWidth / 2;

      let processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, processPortsOrdering);
      processNode.data("justAdded", true);

      const numOfInputNodes = inputs.length;
      const numOfOutputNodes = outputs.length;

      let yPosOfInput =
        processPosition.y -
        ((numOfInputNodes - 1) / 2) *
          (simpleChemicalHeight + tilingPaddingVertical);

      inputs.forEach(function (data, index) {
        const nodeName = data.name;
        const nodeType = data.type;

        if (index === 0) {
          yPosOfInput = processPosition.y;
        } else if (index % 2 === 1) {
          yPosOfInput =
            processPosition.y -
            (simpleChemicalHeight + tilingPaddingVertical) *
              Math.ceil(index / 2);
        } else {
          yPosOfInput =
            processPosition.y +
            (simpleChemicalHeight + tilingPaddingVertical) * (index / 2);
        }

        let nodePosition = {
          x: xPosOfInput,
          y: yPosOfInput,
        };
        if (orientation === "vertical") {
          nodePosition = rotate90(nodePosition, processPosition);
        }

        let newNode = elementUtilities.addNode(nodePosition.x, nodePosition.y, {
          class: nodeType.toLowerCase(),
          language: "PD",
        });
        newNode.data("justAdded", true);
        newNode.data("label", nodeName);

        let newEdge;
        if (reversible) {
          newEdge = elementUtilities.addEdge(
            processNode.id(),
            newNode.id(),
            { class: processLeftSideEdgeType, language: "PD" },
            undefined,
            undefined,
            1
          );
        } else {
          newEdge = elementUtilities.addEdge(newNode.id(), processNode.id(), {
            class: processLeftSideEdgeType,
            language: "PD",
          });
        }
        newEdge.data("justAdded", true);
      });

      let yPosOfOutput =
        processPosition.y -
        ((numOfOutputNodes - 1) / 2) *
          (simpleChemicalHeight + tilingPaddingVertical);

      outputs.forEach(function (data, index) {
        const nodeName = data.name;
        const nodeType = data.type;

        if (index === 0) {
          yPosOfOutput = processPosition.y;
        } else if (index % 2 === 1) {
          yPosOfOutput =
            processPosition.y -
            (simpleChemicalHeight + tilingPaddingVertical) *
              Math.ceil(index / 2);
        } else {
          yPosOfOutput =
            processPosition.y +
            (simpleChemicalHeight + tilingPaddingVertical) * (index / 2);
        }

        let nodePosition = {
          x: xPosOfOutput,
          y: yPosOfOutput,
        };
        if (orientation === "vertical") {
          nodePosition = rotate90(nodePosition, processPosition);
        }

        let newNode = elementUtilities.addNode(nodePosition.x, nodePosition.y, {
          class: nodeType.toLowerCase(),
          language: "PD",
        });
        newNode.data("justAdded", true);
        newNode.data("label", nodeName);

        let newEdge = elementUtilities.addEdge(
          processNode.id(),
          newNode.id(),
          { class: processRightSideEdgeType, language: "PD" },
          undefined,
          undefined,
          0
        );
        newEdge.data("justAdded", true);
      });

      // add regulator node
      if (hasRegulator) {
        const regulatorName = regulator.name;
        const regulatorType = regulator.type;
        let xPosOfRegulator = processPosition.x;
        if (regulatorMultimer.enabled && orientation === "horizontal") {
          xPosOfRegulator -= multimerOffset;
        }
        let yPosOfRegulator =
          processPosition.y -
          (processHeight / 2 + regulatorHeight / 2 + edgeLength);

        let nodePosition = {
          x: xPosOfRegulator,
          y: yPosOfRegulator,
        };
        if (orientation === "vertical") {
          nodePosition = rotate90(nodePosition, processPosition);
        }

        let regulatorNode = elementUtilities.addNode(
          nodePosition.x,
          nodePosition.y,
          { class: regulatorType, language: "PD" }
        );
        regulatorNode.data("justAdded", true);
        regulatorNode.data("label", regulatorName);

        if (regulatorMultimer.enabled) {
          elementUtilities.setMultimerStatus(regulatorNode, true);

          const cardinality = regulatorMultimer.cardinality;
          if (cardinality != "") {
            const infoboxLabel = "N:" + cardinality;
            infoboxObject = {
              clazz: "unit of information",
              label: {
                text: infoboxLabel,
              },
              bbox: {
                w: infoboxLabel.length * 6,
                h: 15,
              },
            };

            elementUtilities.addStateOrInfoBox(regulatorNode, infoboxObject);
          }
        }

        let regulatorEdge = elementUtilities.addEdge(
          regulatorNode.id(),
          processNode.id(),
          { class: "catalysis", language: "PD" }
        );
        regulatorEdge.data("justAdded", true);
      }

      cy.endBatch();

      const eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles;
    };

    elementUtilities.createMetabolicCatalyticActivity = function (
      inputNodeList,
      outputNodeList,
      catalystName,
      catalystType,
      processPosition,
      tilingPaddingVertical,
      tilingPaddingHorizontal,
      edgeLength
    ) {
      var defaultMacromoleculProperties =
        elementUtilities.getDefaultProperties("macromolecule");
      var defaultSimpleChemicalProperties =
        elementUtilities.getDefaultProperties("simple chemical");
      var defaultCatalystTypeProperties =
        elementUtilities.getDefaultProperties(catalystType);
      var defaultProcessProperties =
        elementUtilities.getDefaultProperties("catalytic");
      var processWidth = defaultProcessProperties.width || 50;
      var processHeight = defaultProcessProperties.height || 50;
      var simpleChemicalHeight = defaultSimpleChemicalProperties.height || 35;
      var macromoleculeWidth = defaultMacromoleculProperties.width || 50;
      var macromoleculeHeight = defaultMacromoleculProperties.height || 50;
      var catalystHeight = defaultCatalystTypeProperties.height || 50;
      var processPosition =
        processPosition ||
        elementUtilities.convertToModelPosition({
          x: cy.width() / 2,
          y: cy.height() / 2,
        });
      var tilingPaddingVertical = tilingPaddingVertical || 15;
      var tilingPaddingHorizontal = tilingPaddingHorizontal || 15;
      var edgeLength = edgeLength || 60;

      cy.startBatch();
      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      var xPosOfInput =
        processPosition.x -
        edgeLength -
        processWidth / 2 -
        macromoleculeWidth / 2;
      var xPosOfOutput =
        processPosition.x +
        edgeLength +
        processWidth / 2 +
        macromoleculeWidth / 2;

      var processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, "L-to-R");
      processNode.data("justAdded", true);

      const numOfInputNodes = inputNodeList.length;
      const numOfOutputNodes = outputNodeList.length;
      var yPosOfInput =
        processPosition.y -
        ((numOfInputNodes - 1) / 2) *
          (macromoleculeHeight + tilingPaddingVertical);

      // add input side nodes
      for (var i = 0; i < numOfInputNodes; i++) {
        if (inputNodeList[i].type == "Simple Chemical") {
          var newNode = elementUtilities.addNode(xPosOfInput, yPosOfInput, {
            class: "simple chemical",
            language: "PD",
          });
          yPosOfInput += simpleChemicalHeight + tilingPaddingVertical;
        } else {
          var newNode = elementUtilities.addNode(xPosOfInput, yPosOfInput, {
            class: "macromolecule",
            language: "PD",
          });
          //update the y position
          yPosOfInput += macromoleculeHeight + tilingPaddingVertical;
        }
        newNode.data("justAdded", true);
        newNode.data("label", inputNodeList[i].name);

        var newEdge = elementUtilities.addEdge(newNode.id(), processNode.id(), {
          class: "consumption",
          language: "PD",
        });
        newEdge.data("justAdded", true);
      }

      var yPosOfOutput =
        processPosition.y -
        ((numOfOutputNodes - 1) / 2) *
          (macromoleculeHeight + tilingPaddingVertical);

      // add output side nodes
      for (var i = 0; i < numOfOutputNodes; i++) {
        if (outputNodeList[i].type == "Simple Chemical") {
          var newNode = elementUtilities.addNode(xPosOfOutput, yPosOfOutput, {
            class: "simple chemical",
            language: "PD",
          });
          yPosOfOutput += simpleChemicalHeight + tilingPaddingVertical;
        } else {
          var newNode = elementUtilities.addNode(xPosOfOutput, yPosOfOutput, {
            class: "macromolecule",
            language: "PD",
          });
          //update the y position
          yPosOfOutput += macromoleculeHeight + tilingPaddingVertical;
        }
        newNode.data("justAdded", true);
        newNode.data("label", outputNodeList[i].name);

        var newEdge = elementUtilities.addEdge(processNode.id(), newNode.id(), {
          class: "production",
          language: "PD",
        });
        newEdge.data("justAdded", true);
      }

      // add catalyst node
      var xPosOfCatalyst = processPosition.x;
      var yPosOfCatalyst =
        processPosition.y -
        (processHeight + catalystHeight + tilingPaddingVertical);
      var catalystNode = elementUtilities.addNode(
        xPosOfCatalyst,
        yPosOfCatalyst,
        { class: catalystType, language: "PD" }
      );
      catalystNode.data("justAdded", true);
      catalystNode.data("label", catalystName);

      var catalystEdge = elementUtilities.addEdge(
        catalystNode.id(),
        processNode.id(),
        { class: "catalysis", language: "PD" }
      );
      catalystEdge.data("justAdded", true);

      cy.endBatch();

      //filter the just added elememts to return them and remove just added mark
      var eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles; // Return the just added elements
    };

    elementUtilities.createActivationReaction = function (
      proteinName,
      processPosition,
      edgeLength,
      reverse
    ) {
      var defaultMacromoleculProperties =
        elementUtilities.getDefaultProperties("macromolecule");
      var defaultProcessProperties =
        elementUtilities.getDefaultProperties("activation");
      var processWidth = defaultProcessProperties.width || 50;
      var macromoleculeWidth = defaultMacromoleculProperties.width || 50;
      var processPosition =
        processPosition ||
        elementUtilities.convertToModelPosition({
          x: cy.width() / 2,
          y: cy.height() / 2,
        });
      var edgeLength = edgeLength || 60;

      cy.startBatch();
      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      var xPosOfInput =
        processPosition.x -
        edgeLength -
        processWidth / 2 -
        macromoleculeWidth / 2;
      var xPosOfOutput =
        processPosition.x +
        edgeLength +
        processWidth / 2 +
        macromoleculeWidth / 2;

      var processNode = elementUtilities.addNode(
        processPosition.x,
        processPosition.y,
        { class: "process", language: "PD" }
      );
      elementUtilities.setPortsOrdering(processNode, "L-to-R");
      processNode.data("justAdded", true);

      var yPosition = processPosition.y;

      var inputNode = elementUtilities.addNode(xPosOfInput, yPosition, {
        class: "macromolecule",
        language: "PD",
      });
      inputNode.data("justAdded", true);
      inputNode.data("label", proteinName);
      var infoboxObject = {
        clazz: "unit of information",
        label: {
          text: reverse ? "active" : "inactive",
        },
        style: {
          "shape-name": "ellipse",
        },
        bbox: {
          w: 36,
          h: 15,
        },
      };
      elementUtilities.addStateOrInfoBox(inputNode, infoboxObject);

      var outputNode = elementUtilities.addNode(xPosOfOutput, yPosition, {
        class: "macromolecule",
        language: "PD",
      });
      outputNode.data("justAdded", true);
      outputNode.data("label", proteinName);
      infoboxObject = {
        clazz: "unit of information",
        label: {
          text: reverse ? "inactive" : "active",
        },
        style: {
          "shape-name": "ellipse",
        },
        bbox: {
          w: 36,
          h: 15,
        },
      };
      elementUtilities.addStateOrInfoBox(outputNode, infoboxObject);

      var inputSideEdge = elementUtilities.addEdge(
        inputNode.id(),
        processNode.id(),
        { class: "consumption", language: "PD" }
      );
      inputSideEdge.data("justAdded", true);
      var outputSideEdge = elementUtilities.addEdge(
        processNode.id(),
        outputNode.id(),
        { class: "production", language: "PD" }
      );
      outputSideEdge.data("justAdded", true);
      cy.endBatch();

      //filter the just added elememts to return them and remove just added mark
      var eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles; // Return the just added elements
    };

    /*
     * Creates a template reaction with given parameters. Requires cose-bilkent layout to tile the free macromolecules included
     * in the complex. Parameters are explained below.
     * templateType: The type of the template reaction. It may be 'association', 'dissociation', 'reversible' or 'irreversible'.
     * nodeList: The list of the names and types of molecules which will involve in the reaction.
     * complexName: The name of the complex in the reaction.
     * processPosition: The modal position of the process in the reaction. The default value is the center of the canvas.
     * tilingPaddingVertical: This option will be passed to the cose-bilkent layout with the same name. The default value is 15.
     * tilingPaddingHorizontal: This option will be passed to the cose-bilkent layout with the same name. The default value is 15.
     * edgeLength: The distance between the process and the macromolecules at the both sides.
     */
    elementUtilities.createTemplateReaction = function (
      templateType,
      nodeList,
      complexName,
      processPosition,
      tilingPaddingVertical,
      tilingPaddingHorizontal,
      edgeLength,
      layoutParam
    ) {
      var defaultMacromoleculProperties =
        elementUtilities.getDefaultProperties("macromolecule");
      var defaultSimpleChemicalProperties =
        elementUtilities.getDefaultProperties("simple chemical");
      var defaultProcessProperties =
        elementUtilities.getDefaultProperties(templateType);
      var processWidth = defaultProcessProperties.width || 50;
      var macromoleculeWidth = defaultMacromoleculProperties.width || 50;
      var macromoleculeHeight = defaultMacromoleculProperties.height || 50;
      var simpleChemicalWidth = defaultSimpleChemicalProperties.width || 35;
      var simpleChemicalHeight = defaultSimpleChemicalProperties.height || 35;
      var processPosition =
        processPosition ||
        elementUtilities.convertToModelPosition({
          x: cy.width() / 2,
          y: cy.height() / 2,
        });
      var nodeList = nodeList;
      var complexName = complexName;
      var numOfMolecules = nodeList.length;
      var tilingPaddingVertical = tilingPaddingVertical || 15;
      var tilingPaddingHorizontal = tilingPaddingHorizontal || 15;
      var edgeLength = edgeLength || 60;

      cy.startBatch();

      if (!elementUtilities.getMapType()) {
        elementUtilities.setMapType("PD");
      }

      var xPositionOfFreeMacromolecules;
      var xPositionOfInputMacromolecules;
      if (templateType === "association") {
        xPositionOfFreeMacromolecules =
          processPosition.x -
          edgeLength -
          processWidth / 2 -
          macromoleculeWidth / 2;
      } else if (templateType === "dissociation") {
        xPositionOfFreeMacromolecules =
          processPosition.x +
          edgeLength +
          processWidth / 2 +
          macromoleculeWidth / 2;
      } else {
        xPositionOfFreeMacromolecules =
          processPosition.x -
          edgeLength -
          processWidth / 2 -
          macromoleculeWidth / 2;
        xPositionOfInputMacromolecules =
          processPosition.x +
          edgeLength +
          processWidth / 2 +
          macromoleculeWidth / 2;
      }

      //Create the process in template type
      var process;
      if (templateType === "reversible" || templateType === "irreversible") {
        process = elementUtilities.addNode(
          processPosition.x,
          processPosition.y,
          { class: "process", language: "PD" }
        );
        elementUtilities.setPortsOrdering(process, "L-to-R");
      } else {
        process = elementUtilities.addNode(
          processPosition.x,
          processPosition.y,
          { class: templateType, language: "PD" }
        );
        elementUtilities.setPortsOrdering(process, "L-to-R");
      }
      process.data("justAdded", true);

      //Define the starting y position
      var yPosition =
        processPosition.y -
        ((numOfMolecules - 1) / 2) *
          (macromoleculeHeight + tilingPaddingVertical);

      //Create the free molecules
      for (var i = 0; i < numOfMolecules; i++) {
        // node addition operation is determined by molecule type
        if (nodeList[i].type == "Simple Chemical") {
          var newNode = elementUtilities.addNode(
            xPositionOfFreeMacromolecules,
            yPosition,
            { class: "simple chemical", language: "PD" }
          );
          //update the y position
          yPosition += simpleChemicalHeight + tilingPaddingVertical;
        } else {
          var newNode = elementUtilities.addNode(
            xPositionOfFreeMacromolecules,
            yPosition,
            { class: "macromolecule", language: "PD" }
          );
          //update the y position
          yPosition += macromoleculeHeight + tilingPaddingVertical;
        }
        newNode.data("justAdded", true);
        newNode.data("label", nodeList[i].name);

        //create the edge connected to the new molecule
        var newEdge;
        if (templateType === "association") {
          newEdge = elementUtilities.addEdge(newNode.id(), process.id(), {
            class: "consumption",
            language: "PD",
          });
        } else if (templateType === "dissociation") {
          newEdge = elementUtilities.addEdge(process.id(), newNode.id(), {
            class: "production",
            language: "PD",
          });
        } else {
          //Group right or top elements in group id 1
          if (templateType === "irreversible") {
            newEdge = elementUtilities.addEdge(newNode.id(), process.id(), {
              class: "consumption",
              language: "PD",
            });
          } else {
            newEdge = elementUtilities.addEdge(
              process.id(),
              newNode.id(),
              { class: "production", language: "PD" },
              undefined,
              undefined,
              1
            );
          }
        }

        newEdge.data("justAdded", true);
      }

      if (templateType === "association" || templateType == "dissociation") {
        //Create the complex including macromolecules inside of it
        //Temprorarily add it to the process position we will move it according to the last size of it
        var complex = elementUtilities.addNode(
          processPosition.x,
          processPosition.y,
          { class: "complex", language: "PD" }
        );
        complex.data("justAdded", true);
        complex.data("justAddedLayoutNode", true);

        //If a name is specified for the complex set its label accordingly
        if (complexName) {
          complex.data("label", complexName);
        }

        //create the edge connnected to the complex
        var edgeOfComplex;

        if (templateType === "association") {
          edgeOfComplex = elementUtilities.addEdge(process.id(), complex.id(), {
            class: "production",
            language: "PD",
          });
        } else {
          edgeOfComplex = elementUtilities.addEdge(complex.id(), process.id(), {
            class: "consumption",
            language: "PD",
          });
        }

        edgeOfComplex.data("justAdded", true);

        for (var i = 0; i < numOfMolecules; i++) {
          // Add a molecule(dependent on it's type) not having a previously defined id and having the complex created in this reaction as parent
          if (nodeList[i].type == "Simple Chemical") {
            var newNode = elementUtilities.addNode(
              complex.position("x"),
              complex.position("y"),
              { class: "simple chemical", language: "PD" },
              undefined,
              complex.id()
            );
          } else {
            var newNode = elementUtilities.addNode(
              complex.position("x"),
              complex.position("y"),
              { class: "macromolecule", language: "PD" },
              undefined,
              complex.id()
            );
          }

          newNode.data("justAdded", true);
          newNode.data("label", nodeList[i].name);
          newNode.data("justAddedLayoutNode", true);
        }
      } else {
        //Create the input macromolecules
        var numOfInputMacromolecules = complexName.length;
        yPosition =
          processPosition.y -
          ((numOfInputMacromolecules - 1) / 2) *
            (macromoleculeHeight + tilingPaddingVertical);

        for (var i = 0; i < numOfInputMacromolecules; i++) {
          if (complexName[i].type == "Simple Chemical") {
            var newNode = elementUtilities.addNode(
              xPositionOfInputMacromolecules,
              yPosition,
              { class: "simple chemical", language: "PD" }
            );
            yPosition += simpleChemicalHeight + tilingPaddingVertical;
          } else {
            var newNode = elementUtilities.addNode(
              xPositionOfInputMacromolecules,
              yPosition,
              { class: "macromolecule", language: "PD" }
            );
            yPosition += macromoleculeHeight + tilingPaddingVertical;
          }

          newNode.data("justAdded", true);
          newNode.data("label", complexName[i].name);

          //create the edge connected to the new macromolecule
          var newEdge;

          //Group the left or bottom elements in group id 0 if reversible
          if (templateType === "irreversible") {
            newEdge = elementUtilities.addEdge(process.id(), newNode.id(), {
              class: "production",
              language: "PD",
            });
          } else {
            newEdge = elementUtilities.addEdge(
              process.id(),
              newNode.id(),
              { class: "production", language: "PD" },
              undefined,
              undefined,
              0
            );
          }
          newEdge.data("justAdded", true);
        }
      }

      cy.endBatch();

      var layoutNodes = cy.nodes("[justAddedLayoutNode]");
      layoutNodes.removeData("justAddedLayoutNode");
      var layout = layoutNodes.layout({
        name: layoutParam.name,
        randomize: false,
        fit: false,
        animate: false,
        tilingPaddingVertical: tilingPaddingVertical,
        tilingPaddingHorizontal: tilingPaddingHorizontal,
        stop: function () {
          //If it is a reversible reaction no need to re-position complexes
          if (templateType === "reversible") return;
          //re-position the nodes inside the complex
          var supposedXPosition;
          var supposedYPosition = processPosition.y;

          if (templateType === "association") {
            supposedXPosition =
              processPosition.x +
              edgeLength +
              processWidth / 2 +
              complex.outerWidth() / 2;
          } else {
            supposedXPosition =
              processPosition.x -
              edgeLength -
              processWidth / 2 -
              complex.outerWidth() / 2;
          }

          var positionDiffX = (supposedXPosition - complex.position("x")) / 2;
          var positionDiffY = (supposedYPosition - complex.position("y")) / 2;
          elementUtilities.moveNodes(
            { x: positionDiffX, y: positionDiffY },
            complex
          );
        },
      });

      // Do this check for cytoscape.js backward compatibility
      if (
        layout &&
        layout.run &&
        templateType !== "reversible" &&
        templateType !== "irreversible"
      ) {
        layout.run();
      }

      //filter the just added elememts to return them and remove just added mark
      var eles = cy.elements("[justAdded]");
      eles.removeData("justAdded");

      cy.elements().unselect();
      eles.select();

      return eles; // Return the just added elements
    };

    /*
     * Move the nodes to a new parent and change their position if possDiff params are set.
     */
    elementUtilities.changeParent = function (
      nodes,
      newParent,
      posDiffX,
      posDiffY
    ) {
      var newParentId =
        newParent == undefined || typeof newParent === "string"
          ? newParent
          : newParent.id();
      var movedEles = nodes.move({ parent: newParentId });
      if (typeof posDiffX != "undefined" || typeof posDiffY != "undefined") {
        elementUtilities.moveNodes({ x: posDiffX, y: posDiffY }, nodes);
      }
      elementUtilities.maintainPointer(movedEles);
      return movedEles;
    };

    elementUtilities.updateInfoboxStyle = function (node, index, newProps) {
      var infoboxObj = node.data("statesandinfos")[index];
      $.extend(infoboxObj.style, newProps);
      cy.style().update();
    };

    elementUtilities.updateInfoboxObj = function (node, index, newProps) {
      var infoboxObj = node.data("statesandinfos")[index];
      $.extend(infoboxObj, newProps);
    };

    // Resize given nodes if useAspectRatio is truthy one of width or height should not be set.
    elementUtilities.resizeNodes = function (
      nodes,
      width,
      height,
      useAspectRatio,
      preserveRelativePos
    ) {
      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var ratio = undefined;
        var eleMustBeSquare = elementUtilities.mustBeSquare(node.data("class"));

        if (preserveRelativePos === true) {
          var oldWidth = node.data("bbox").w;
          var oldHeight = node.data("bbox").h;
        }

        // Note that both width and height should not be set if useAspectRatio is truthy
        if (!node.isParent()) {
          if (width) {
            if (useAspectRatio || eleMustBeSquare) {
              ratio = width / node.width();
            }

            node.data("bbox").w = width;
          }

          if (height) {
            if (useAspectRatio || eleMustBeSquare) {
              ratio = height / node.height();
            }

            node.data("bbox").h = height;
          }

          if (ratio && !height) {
            node.data("bbox").h = node.height() * ratio;
          } else if (ratio && !width) {
            node.data("bbox").w = node.width() * ratio;
          }
        } else {
          node.data("minHeight", "" + height);
          node.data("minWidth", "" + width);
          node.data("minWidthBiasLeft", "50%");
          node.data("minWidthBiasRight", "50%");
          node.data("minHeightBiasTop", "50%");
          node.data("minHeightBiasBottom", "50%");
        }

        /*    if (preserveRelativePos === true) {
          var statesandinfos = node.data('statesandinfos');
          var topBottom = statesandinfos.filter(box => (box.anchorSide === "top" || box.anchorSide === "bottom"));
          var rightLeft = statesandinfos.filter(box => (box.anchorSide === "right" || box.anchorSide === "left"));

          topBottom.forEach(function(box){
            if (box.bbox.x < 0) {
              box.bbox.x = 0;
            }
            else if (box.bbox.x > oldWidth) {
              box.bbox.x = oldWidth;
            }
            box.bbox.x = node.data("bbox").w * box.bbox.x / oldWidth;
          });

          rightLeft.forEach(function(box){
            if (box.bbox.y < 0) {
              box.bbox.y = 0;
            }
            else if (box.bbox.y > oldHeight) {
              box.bbox.y = oldHeight;
            }
            box.bbox.y = node.data("bbox").h * box.bbox.y / oldHeight;
          });
        } */
      }
    };

    elementUtilities.calculateMinWidth = function (node) {
      var defaultWidth = this.getDefaultProperties(node.data("class")).width;

      // Label width calculation
      var style = node.style();

      var fontFamiliy = style["font-family"];
      var fontSize = style["font-size"];
      var labelText = style["label"];

      if (labelText === "" && node.data("label") && node.data("label") !== "") {
        labelText = node.data("label");
      }

      var labelWidth = elementUtilities.getWidthByContent(
        labelText,
        fontFamiliy,
        fontSize
      );

      var statesandinfos = node.data("statesandinfos");
      //Top and bottom infoBoxes
      //var topInfoBoxes = statesandinfos.filter(box => (box.anchorSide === "top" || ((box.anchorSide === "right" || box.anchorSide === "left") && (box.bbox.y <= 12))));
      //var bottomInfoBoxes = statesandinfos.filter(box => (box.anchorSide === "bottom" || ((box.anchorSide === "right" || box.anchorSide === "left") && (box.bbox.y >= node.data('bbox').h - 12))));
      var unitGap = 5;
      var topIdealWidth = unitGap;
      var bottomIdealWidth = unitGap;
      var rightMaxWidth = 0;
      var leftMaxWidth = 0;
      statesandinfos.forEach(function (box) {
        if (box.anchorSide === "top") {
          topIdealWidth += box.bbox.w + unitGap;
        } else if (box.anchorSide === "bottom") {
          bottomIdealWidth += box.bbox.w + unitGap;
        } else if (box.anchorSide === "right") {
          rightMaxWidth =
            box.bbox.w > rightMaxWidth ? box.bbox.w : rightMaxWidth;
        } else {
          leftMaxWidth = box.bbox.w > leftMaxWidth ? box.bbox.w : leftMaxWidth;
        }
      });

      var middleWidth =
        labelWidth + 2 * Math.max(rightMaxWidth / 2, leftMaxWidth / 2);

      var compoundWidth = 0;
      if (node.isParent()) {
        compoundWidth = node.children().boundingBox().w;
      }
      return Math.max(
        middleWidth,
        defaultWidth / 2,
        topIdealWidth,
        bottomIdealWidth,
        compoundWidth
      );
    };

    elementUtilities.calculateMinHeight = function (node) {
      var statesandinfos = node.data("statesandinfos");
      var margin = 7;
      var unitGap = 5;
      var defaultHeight = this.getDefaultProperties(node.data("class")).height;
      var leftInfoBoxes = statesandinfos.filter(
        (box) => box.anchorSide === "left"
      );
      var leftHeight = unitGap;
      leftInfoBoxes.forEach(function (box) {
        leftHeight += box.bbox.h + unitGap;
      });
      var rightInfoBoxes = statesandinfos.filter(
        (box) => box.anchorSide === "right"
      );
      var rightHeight = unitGap;
      rightInfoBoxes.forEach(function (box) {
        rightHeight += box.bbox.h + unitGap;
      });
      var style = node.style();
      var labelText = style["label"].split("\n").filter((text) => text !== "");
      var fontSize = parseFloat(
        style["font-size"].substring(0, style["font-size"].length - 2)
      );
      var totalHeight = labelText.length * fontSize + 2 * margin;

      var compoundHeight = 0;
      if (node.isParent()) {
        compoundHeight = node.children().boundingBox().h;
      }
      return Math.max(
        totalHeight,
        defaultHeight / 2,
        leftHeight,
        rightHeight,
        compoundHeight
      );
    };

    elementUtilities.isResizedToContent = function (node) {
      if (!node || !node.isNode() || !node.data("bbox")) {
        return false;
      }

      //var w = node.data('bbox').w;
      //var h = node.data('bbox').h;
      var w = node.width();
      var h = node.height();

      var minW = elementUtilities.calculateMinWidth(node);
      var minH = elementUtilities.calculateMinHeight(node);

      if (w === minW && h === minH) return true;
      else return false;
    };

    // Section End
    // Add remove utilities

    // Relocates state and info boxes. This function is expected to be called after add/remove state and info boxes
    elementUtilities.relocateStateAndInfos = function (ele) {
      var stateAndInfos =
        ele.isNode && ele.isNode() ? ele.data("statesandinfos") : ele;
      var length = stateAndInfos.length;
      if (length == 0) {
        return;
      } else if (length == 1) {
        stateAndInfos[0].bbox.x = 0;
        stateAndInfos[0].bbox.y = -50;
      } else if (length == 2) {
        stateAndInfos[0].bbox.x = 0;
        stateAndInfos[0].bbox.y = -50;

        stateAndInfos[1].bbox.x = 0;
        stateAndInfos[1].bbox.y = 50;
      } else if (length == 3) {
        stateAndInfos[0].bbox.x = -25;
        stateAndInfos[0].bbox.y = -50;

        stateAndInfos[1].bbox.x = 25;
        stateAndInfos[1].bbox.y = -50;

        stateAndInfos[2].bbox.x = 0;
        stateAndInfos[2].bbox.y = 50;
      } else {
        stateAndInfos[0].bbox.x = -25;
        stateAndInfos[0].bbox.y = -50;

        stateAndInfos[1].bbox.x = 25;
        stateAndInfos[1].bbox.y = -50;

        stateAndInfos[2].bbox.x = -25;
        stateAndInfos[2].bbox.y = 50;

        stateAndInfos[3].bbox.x = 25;
        stateAndInfos[3].bbox.y = 50;
      }
    };

    // Change state value or unit of information box of given nodes with given index.
    // Type parameter indicates whether to change value or variable, it is valid if the box at the given index is a state variable.
    // Value parameter is the new value to set.
    // This method returns the old value of the changed data (We assume that the old value of the changed data was the same for all nodes).
    // Each character assumed to occupy 8 unit
    // Each infobox can have at most 32 units of width
    elementUtilities.changeStateOrInfoBox = function (
      nodes,
      index,
      value,
      type
    ) {
      var result;
      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var sbgnclass = node.data("class");
        var stateAndInfos = node.data("statesandinfos");
        var box = stateAndInfos[index];
        var oldLength = box.bbox.w;
        var newLength = 0;

        var content = "";
        if (box.clazz == "state variable") {
          if (!result) {
            result = box.state[type];
          }

          box.state[type] = value;
          if (box.state["value"] !== undefined) {
            content += box.state["value"];
          }
          if (
            box.state["variable"] !== undefined &&
            box.state["variable"].length > 0
          ) {
            content += box.state["variable"] + "@";
          }
        } else if (box.clazz == "unit of information") {
          if (!result) {
            result = box.label.text;
          }
          content += value;
          box.label.text = value;
        } else if (box.clazz == "residue variable") {
          if (!result) {
            result = box.residue[type];
          }

          box.residue[type] = value;
          if (
            box.residue["variable"] !== undefined &&
            box.residue["variable"].length > 0
          ) {
            content += box.residue["variable"];
          }
        } else if (box.clazz == "binding region") {
          if (!result) {
            result = box.region[type];
          }

          box.region[type] = value;
          if (
            box.region["variable"] !== undefined &&
            box.region["variable"].length > 0
          ) {
            content += box.region["variable"];
          }
        }

        var min =
          sbgnclass === "SIF macromolecule" ||
          sbgnclass === "SIF simple chemical"
            ? 15
            : 12;
        var fontFamily = box.style["font-family"];
        var fontSize = box.style["font-size"];
        var borderWidth = box.style["border-width"];
        var opts = {
          min,
          max: 48,
          margin: borderWidth / 2 + 0.5,
        };
        var previousWidth = box.bbox.w;
        box.bbox.w = elementUtilities.getWidthByContent(
          content,
          fontFamily,
          fontSize,
          opts
        );

        if (box.anchorSide == "top" || box.anchorSide == "bottom") {
          var unitLayout = node.data()["auxunitlayouts"][box.anchorSide];
          if (unitLayout.units[unitLayout.units.length - 1].id == box.id) {
            var borderWidth = node.data()["border-width"];
            var shiftAmount =
              (((box.bbox.w - previousWidth) / 2) * 100) /
              (node.outerWidth() - borderWidth);

            if (shiftAmount >= 0) {
              if (box.bbox.x + shiftAmount <= 100) {
                box.bbox.x = box.bbox.x + shiftAmount;
              }
            }
            /*  else{
              var previousInfoBbox = {x : 0, w:0};
              if(unitLayout.units.length > 1){
                previousInfoBbox= unitLayout.units[unitLayout.units.length-2].bbox;      
              }

              
              
              sbgnvizInstance.classes.AuxUnitLayout.setIdealGap(node, box.anchorSide);
              var idealGap = sbgnvizInstance.classes.AuxUnitLayout.getCurrentGap(box.anchorSide);
              var newPosition = previousInfoBbox.x + (previousInfoBbox.w/2 + idealGap + box.bbox.w/2)*100 / (node.outerWidth() - borderWidth);
              box.bbox.x = newPosition;
              
            } */
          }
        }

        /* if (box.anchorSide === "top" || box.anchorSide === "bottom") {
          box.bbox.x += (box.bbox.w - oldLength) / 2;
          var units = (node.data('auxunitlayouts')[box.anchorSide]).units;
          var shiftIndex = 0;
          for (var i = 0; i < units.length; i++) {
            if(units[i] === box){
              shiftIndex = i;
              break;
            }
          }
          for (var j = shiftIndex+1; j < units.length; j++) {
              units[j].bbox.x += (box.bbox.w - oldLength);
          }
        } */
      }

      //TODO find a way to elimate this redundancy to update info-box positions
      node.data("border-width", node.data("border-width"));

      return result;
    };

    // Add a new state or info box to given nodes.
    // The box is represented by the parameter obj.
    // This method returns the index of the just added box.
    elementUtilities.addStateOrInfoBox = function (nodes, obj) {
      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var locationObj;

        var defaultProps = elementUtilities.getDefaultProperties(
          node.data("class")
        );
        var infoboxProps = defaultProps[obj.clazz];
        var bbox = obj.bbox || {
          w: infoboxProps.width,
          h: infoboxProps.height,
        };
        var style = elementUtilities.getDefaultInfoboxStyle(
          node.data("class"),
          obj.clazz
        );
        if (obj.style) {
          $.extend(style, obj.style);
        }

        if (obj.clazz == "unit of information") {
          locationObj = sbgnvizInstance.classes.UnitOfInformation.create(
            node,
            cy,
            obj.label.text,
            bbox,
            obj.location,
            obj.position,
            style,
            obj.index,
            obj.id
          );
        } else if (obj.clazz == "state variable") {
          locationObj = sbgnvizInstance.classes.StateVariable.create(
            node,
            cy,
            obj.state.value,
            obj.state.variable,
            bbox,
            obj.location,
            obj.position,
            style,
            obj.index,
            obj.id
          );
        } else if (obj.clazz == "residue variable") {
          locationObj = sbgnvizInstance.classes.ResidueVariable.create(
            node,
            cy,
            obj.residue.value,
            obj.residue.variable,
            bbox,
            obj.location,
            obj.position,
            style,
            obj.index,
            obj.id
          );
        } else if (obj.clazz == "binding region") {
          locationObj = sbgnvizInstance.classes.BindingRegion.create(
            node,
            cy,
            obj.region.value,
            obj.region.variable,
            bbox,
            obj.location,
            obj.position,
            style,
            obj.index,
            obj.id
          );
        }
      }
      return locationObj;
    };

    // Remove the state or info boxes of the given nodes at given index.
    // Returns the removed box.
    elementUtilities.removeStateOrInfoBox = function (nodes, locationObj) {
      var obj;
      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var stateAndInfos = node.data("statesandinfos");
        var unit = stateAndInfos[locationObj.index];

        var unitClass = sbgnvizInstance.classes.getAuxUnitClass(unit);

        obj = unitClass.remove(unit, cy);
      }

      return obj;
    };

    //Tiles informations boxes for given anchorSides
    elementUtilities.fitUnits = function (node, locations) {
      var obj = [];
      node.data("statesandinfos").forEach(function (ele) {
        obj.push({
          x: ele.bbox.x,
          y: ele.bbox.y,
          anchorSide: ele.anchorSide,
        });
      });
      sbgnvizInstance.classes.AuxUnitLayout.fitUnits(node, cy, locations);
      return obj;
    };

    //Check which anchorsides fits
    elementUtilities.checkFit = function (node, location) {
      //if no location given, it checks all possible locations
      console.log("elementUtilities.checkFit", node);
      return sbgnvizInstance.classes.AuxUnitLayout.checkFit(node, cy, location);
    };

    //Modify array of aux layout units
    elementUtilities.modifyUnits = function (node, unit, anchorSide) {
      sbgnvizInstance.classes.AuxUnitLayout.modifyUnits(
        node,
        unit,
        anchorSide,
        cy
      );
    };

    // Set multimer status of the given nodes to the given status.
    elementUtilities.setMultimerStatus = function (nodes, status) {
      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var sbgnclass = node.data("class");
        var isMultimer = node.data("class").endsWith(" multimer");

        if (status) {
          // Make multimer status true
          if (!isMultimer) {
            node.data("class", sbgnclass + " multimer");
          }
        } else {
          // Make multimer status false
          if (isMultimer) {
            node.data("class", sbgnclass.replace(" multimer", ""));
          }
        }
      }
    };

    //Need to add this to doc : TO-DO
    elementUtilities.setActiveStatus = function (nodes, status) {
      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var sbgnclass = node.data("class");
        var isActive = node.data("class").startsWith("active ");

        if (status) {
          // Make multimer status true
          if (!isActive) {
            node.data("class", "active " + sbgnclass);
            //node.data('class', sbgnclass + ' multimer');
          }
        } else {
          // Make multimer status false
          if (isActive) {
            node.data("class", sbgnclass.replace("active ", ""));
            //node.data('class', sbgnclass.replace(' multimer', ''));
          }
        }
      }
    };

    //Need to add this to doc : TO-DO
    elementUtilities.setHypotheticalStatus = function (nodes, status) {
      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var sbgnclass = node.data("class");
        var isHypothetical = node.data("class").includes("hypothetical");
        var isActive = node.data("class").startsWith("active ");

        if (status) {
          // Make multimer status true
          if (!isHypothetical) {
            if (isActive) {
              var tmp = sbgnclass.substring(7);
              node.data("class", "active hypothetical " + tmp);
            } else {
              node.data("class", "hypothetical " + sbgnclass);
            }
            //node.data('class', sbgnclass + ' multimer');
          }
        } else {
          // Make multimer status false
          if (isHypothetical) {
            node.data("class", sbgnclass.replace("hypothetical ", ""));
            //node.data('class', sbgnclass.replace(' multimer', ''));
          }
        }
      }
    };

    // Change font properties of the given elements with given font data
    elementUtilities.changeFontProperties = function (eles, data) {
      for (var prop in data) {
        eles.data(prop, data[prop]);
      }
    };

    // This function gets an edge, and ends of that edge (Optionally it may take just the classes of the edge as well) as parameters.
    // It may return 'valid' (that ends is valid for that edge), 'reverse' (that ends is not valid for that edge but they would be valid
    // if you reverse the source and target), 'invalid' (that ends are totally invalid for that edge).
    elementUtilities.validateArrowEnds = function (
      edge,
      source,
      target,
      isReplacement
    ) {
      // if map type is Unknown -- no rules applied
      if (
        elementUtilities.getMapType() == "HybridAny" ||
        elementUtilities.getMapType() == "HybridSbgn" ||
        !elementUtilities.getMapType()
      )
        return "valid";

      var edgeclass = typeof edge === "string" ? edge : edge.data("class");
      var sourceclass = source.data("class");
      var targetclass = target.data("class");
      var mapType = elementUtilities.getMapType();
      var edgeConstraints =
        elementUtilities[mapType].connectivityConstraints[edgeclass];
      if (mapType == "AF") {
        if (sourceclass.startsWith("BA"))
          // we have separate classes for each biological activity
          sourceclass = "biological activity"; // but same rule applies to all of them

        if (targetclass.startsWith("BA"))
          // we have separate classes for each biological activity
          targetclass = "biological activity"; // but same rule applies to all of them
      } else if (mapType == "PD") {
        sourceclass = sourceclass.replace(/\s*multimer$/, "");
        targetclass = targetclass.replace(/\s*multimer$/, "");
      } else if (mapType == "SBML") {
        sourceclass = sourceclass.replace(/\s*multimer$/, "");
        targetclass = targetclass.replace(/\s*multimer$/, "");
        sourceclass = sourceclass.replace("active ", "");
        targetclass = targetclass.replace("active ", "");
        sourceclass = sourceclass.replace("hypothetical ", "");
        targetclass = targetclass.replace("hypothetical ", "");
      }

      // given a node, acting as source or target, returns boolean wether or not it has too many edges already
      function hasTooManyEdges(node, sourceOrTarget) {
        var nodeclass = node.data("class");
        nodeclass = nodeclass.replace(/\s*multimer$/, "");
        nodeclass = nodeclass.replace("active ", "");
        nodeclass = nodeclass.replace("hypothetical ", "");
        if (nodeclass.startsWith("BA")) nodeclass = "biological activity";

        /*
          On the logic below:

          Current edge count (incoming or outgoing) of nodes should be strictly less 
          than the maximum allowed if we are adding an edge to the node. This way
          it will never exceed the max count.
          
          Edges can be added in two different ways. Either they are added directly or
          they are added by being replaced from another node, i.e disconnected from
          one and connected to another.

          We can detect if the edge being added is added from a replacement by checking
          whether the source stayed the same when checking edge counts of the source node,
          and whether the target stayed the same when checking edge counts of the
          target node.

          Current edge count of nodes can be allowed to be equal to the maximum in 
          cases where a replacement is made. But we should be careful that this
          replacement operation is not also an addition operation as described above.
        */

        var totalTooMany = true;
        var edgeTooMany = true;
        if (sourceOrTarget == "source") {
          var sameEdgeCountOut = node
            .outgoers('edge[class="' + edgeclass + '"]')
            .size();
          var totalEdgeCountOut = node.outgoers("edge").size();
          var maxTotal = edgeConstraints[nodeclass].asSource.maxTotal;
          var maxEdge = edgeConstraints[nodeclass].asSource.maxEdge;

          var compareStrict = !(isReplacement && edge.source() === source);

          var withinLimits =
            !maxTotal ||
            (compareStrict && totalEdgeCountOut < maxTotal) ||
            (!compareStrict && totalEdgeCountOut <= maxTotal);

          if (withinLimits) {
            totalTooMany = false;
          }
          // then check limits for this specific edge class

          withinLimits =
            !maxEdge ||
            (compareStrict && sameEdgeCountOut < maxEdge) ||
            (!compareStrict && sameEdgeCountOut <= maxEdge);

          if (withinLimits) {
            edgeTooMany = false;
          }

          // if only one of the limits is reached then edge is invalid
          return totalTooMany || edgeTooMany;
        } else {
          // node is used as target
          var sameEdgeCountIn = node
            .incomers('edge[class="' + edgeclass + '"]')
            .size();
          var totalEdgeCountIn = node.incomers("edge").size();
          var maxTotal = edgeConstraints[nodeclass].asTarget.maxTotal;
          var maxEdge = edgeConstraints[nodeclass].asTarget.maxEdge;

          var compareStrict = !(isReplacement && edge.target() === target);

          var withinLimits =
            !maxTotal ||
            (compareStrict && totalEdgeCountIn < maxTotal) ||
            (!compareStrict && totalEdgeCountIn <= maxTotal);

          if (withinLimits) {
            totalTooMany = false;
          }

          withinLimits =
            !maxEdge ||
            (compareStrict && sameEdgeCountIn < maxEdge) ||
            (!compareStrict && sameEdgeCountIn <= maxEdge);

          if (withinLimits) {
            edgeTooMany = false;
          }
          return totalTooMany || edgeTooMany;
        }
      }

      function isInComplex(node) {
        var parentClass = node.parent().data("class");
        return parentClass && parentClass.startsWith("complex");
      }

      if (isInComplex(source) || isInComplex(target)) {
        // subunits of a complex are no longer EPNs, no connection allowed
        return "invalid";
      }

      // check nature of connection
      if (
        edgeConstraints[sourceclass].asSource.isAllowed &&
        edgeConstraints[targetclass].asTarget.isAllowed
      ) {
        // check amount of connections
        if (
          !hasTooManyEdges(source, "source") &&
          !hasTooManyEdges(target, "target")
        ) {
          return "valid";
        }
      }
      // try to reverse
      if (
        edgeConstraints[targetclass].asSource.isAllowed &&
        edgeConstraints[sourceclass].asTarget.isAllowed
      ) {
        if (
          !hasTooManyEdges(target, "source") &&
          !hasTooManyEdges(source, "target")
        ) {
          return "reverse";
        }
      }
      return "invalid";
    };

    elementUtilities.deleteAndPerformLayout = function (eles, layoutparam) {
      var result = eles.remove();
      if (typeof layoutparam === "function") {
        layoutparam(); // If layoutparam is a function execute it
      } else {
        var layout = cy.layout(layoutparam); // If layoutparam is layout options call layout with that options.

        // Do this check for cytoscape.js backward compatibility
        if (layout && layout.run) {
          layout.run();
        }
      }

      return result;
    };

    /*
     * Hide given eles and perform given layout afterward. Layout parameter may be layout options
     * or a function to call.
     */
    elementUtilities.hideAndPerformLayout = function (eles, layoutparam) {
      var result = cy.viewUtilities().hide(eles); // Hide given eles
      if (typeof layoutparam === "function") {
        layoutparam(); // If layoutparam is a function execute it
      } else {
        var layout = cy.layout(layoutparam); // If layoutparam is layout options call layout with that options.

        // Do this check for cytoscape.js backward compatibility
        if (layout && layout.run) {
          layout.run();
        }
      }

      return result;
    };

    /*
     * Unhide given eles and perform given layout afterward. Layout parameter may be layout options
     * or a function to call.
     */
    elementUtilities.showAndPerformLayout = function (eles, layoutparam) {
      var result = cy.viewUtilities().show(eles); // Show given eles
      if (typeof layoutparam === "function") {
        layoutparam(); // If layoutparam is a function execute it
      } else {
        var layout = cy.layout(layoutparam); // If layoutparam is layout options call layout with that options.

        // Do this check for cytoscape.js backward compatibility
        if (layout && layout.run) {
          layout.run();
        }
      }

      return result;
    };

    /*
     * Change style/css of given eles by setting getting property name to the given value/values (Note that valueMap parameter may be
     * a single string or an id to value map).
     */
    elementUtilities.changeCss = function (eles, name, valueMap) {
      if (typeof valueMap === "object") {
        cy.startBatch();
        for (var i = 0; i < eles.length; i++) {
          var ele = cy.getElementById(eles[i].id());
          ele.css(name, valueMap[ele.id()]); // valueMap is an id to value map use it in this way
        }
        cy.endBatch();
      } else {
        eles.css(name, valueMap); // valueMap is just a string set css('name') for all eles to this value
      }
    };

    /*
     * Change data of given eles by setting getting property name to the given value/values (Note that valueMap parameter may be
     * a single string or an id to value map).
     */
    elementUtilities.changeData = function (eles, name, valueMap) {
      console.log("changing data in chise");
      if (typeof valueMap === "object") {
        cy.startBatch();
        for (var i = 0; i < eles.length; i++) {
          var ele = cy.getElementById(eles[i].id());
          ele.data(name, valueMap[ele.id()]); // valueMap is an id to value map use it in this way
        }
        cy.endBatch();
      } else {
        eles.data(name, valueMap); // valueMap is just a string set css('name') for all eles to this value
      }
    };

    elementUtilities.updateSetField = function (
      ele,
      fieldName,
      toDelete,
      toAdd,
      callback
    ) {
      var set = ele.data(fieldName);
      if (!set) {
        return;
      }
      var updates = {};

      if (toDelete != null && set[toDelete]) {
        delete set[toDelete];
        updates.deleted = toDelete;
      }

      if (toAdd != null) {
        set[toAdd] = true;
        updates.added = toAdd;
      }

      if (
        callback &&
        (updates["deleted"] != null || updates["added"] != null)
      ) {
        callback();
      }

      return updates;
    };

    /*
     * Return the set of all nodes present under the given position
     * renderedPos must be a point defined relatively to cytoscape container
     * (like renderedPosition field of a node)
     */
    elementUtilities.getNodesAt = function (renderedPos) {
      var nodes = cy.nodes();
      var x = renderedPos.x;
      var y = renderedPos.y;
      var resultNodes = [];
      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var renderedBbox = node.renderedBoundingBox({
          includeNodes: true,
          includeEdges: false,
          includeLabels: false,
          includeShadows: false,
        });
        if (x >= renderedBbox.x1 && x <= renderedBbox.x2) {
          if (y >= renderedBbox.y1 && y <= renderedBbox.y2) {
            resultNodes.push(node);
          }
        }
      }
      return resultNodes;
    };

    elementUtilities.demultimerizeClass = function (sbgnclass) {
      return sbgnclass.replace(" multimer", "");
    };

    /**
     * @param mapType - type of the current map (PD, AF or Unknown)
     */
    elementUtilities.setMapType = function (mapType) {
      elementUtilities.mapType = mapType;
      return mapType;
    };

    /**
     * return - map type
     */
    elementUtilities.getMapType = function () {
      return elementUtilities.mapType;
    };
    /**
     * Resets map type
     */
    elementUtilities.resetMapType = function () {
      elementUtilities.mapType = undefined;
    };

    /**
     * Keep consistency of links to self inside the data() structure.
     * This is needed whenever a node changes parents, for example,
     * as it is destroyed and recreated. But the data() stays identical.
     * This creates inconsistencies for the pointers stored in data(),
     * as they now point to a deleted node.
     */
    elementUtilities.maintainPointer = function (eles) {
      eles.nodes().forEach(function (ele) {
        // restore background images
        ele.emit("data");

        // skip nodes without any auxiliary units
        if (
          !ele.data("statesandinfos") ||
          ele.data("statesandinfos").length == 0
        ) {
          return;
        }
        for (var side in ele.data("auxunitlayouts")) {
          ele.data("auxunitlayouts")[side].parentNode = ele.id();
        }
        for (var i = 0; i < ele.data("statesandinfos").length; i++) {
          ele.data("statesandinfos")[i].parent = ele.id();
        }
      });
    };

    elementUtilities.anyHasBackgroundImage = function (eles) {
      var obj = elementUtilities.getBackgroundImageObjs(eles);
      if (obj === undefined) return false;
      else {
        for (var key in obj) {
          var value = obj[key];
          if (value && !$.isEmptyObject(value)) return true;
        }
        return false;
      }
    };

    elementUtilities.hasBackgroundImage = function (ele) {
      if (!ele.isNode() || !ele.data("background-image")) {
        return false;
      }
      var bg;

      if (typeof ele.data("background-image") === "string") {
        bg = ele.data("background-image").split(" ");
      } else if (Array.isArray(obj["background-image"])) {
        bg = ele.data("background-image");
      }

      if (!bg) return false;

      var cloneImg =
        "data:image/svg+xml;utf8,%3Csvg%20width%3D%22100%22%20height%3D%22100%22%20viewBox%3D%220%200%20100%20100%22%20style%3D%22fill%3Anone%3Bstroke%3Ablack%3Bstroke-width%3A0%3B%22%20xmlns%3D%22http%3A//www.w3.org/2000/svg%22%20%3E%3Crect%20x%3D%220%22%20y%3D%220%22%20width%3D%22100%22%20height%3D%22100%22%20style%3D%22fill%3A%23838383%22/%3E%20%3C/svg%3E";
      // If cloneImg is not the only image or there are multiple images there is a background image
      var onlyHasCloneMarkerAsBgImage =
        bg.length === 1 && bg.indexOf(cloneImg) === 0;

      if (bg.length > 1 || !onlyHasCloneMarkerAsBgImage) return true;

      return false;
    };

    elementUtilities.getBackgroundImageURL = function (eles) {
      ("getting background images");
      if (!eles || eles.length < 1) return;

      var commonURL = "";
      for (var i = 0; i < eles.length; i++) {
        var ele = eles[i];

        if (!ele.isNode() || !elementUtilities.hasBackgroundImage(ele)) return;

        var url = ele.data("background-image").split(" ").pop();
        if (
          !url ||
          url.indexOf("http") !== 0 ||
          (commonURL !== "" && commonURL !== url)
        )
          return;
        else if (commonURL === "") commonURL = url;
      }

      return commonURL;
    };

    elementUtilities.getBackgroundImageObjs = function (eles) {
      if (!eles || eles.length < 1) return;

      var list = {};
      for (var i = 0; i < eles.length; i++) {
        var ele = eles[i];
        var obj = getBgObj(ele);
        if (Object.keys(obj).length < 1) return;

        list[ele.data("id")] = obj;
      }
      return list;

      function getBgObj(ele) {
        if (ele.isNode() && elementUtilities.hasBackgroundImage(ele)) {
          var keys = [
            "background-image",
            "background-fit",
            "background-image-opacity",
            "background-position-x",
            "background-position-y",
            "background-height",
            "background-width",
          ];

          var obj = {};
          keys.forEach(function (key) {
            var value;
            if (ele.data(key) && typeof ele.data(key) === "string") {
              value = ele.data(key).split(" ")[0];
            } else {
              value = ele.data(key);
            }
            obj[key] = value;
          });

          return obj;
        } else if (ele.isNode()) return {};
      }
    };

    elementUtilities.getBackgroundFitOptions = function (eles) {
      if (!eles || eles.length < 1) return;

      var commonFit = "";
      for (var i = 0; i < eles.length; i++) {
        var node = eles[i];
        if (!node.isNode()) return;

        var fit = getFitOption(node);
        if (!fit || (commonFit !== "" && fit !== commonFit)) return;
        else if (commonFit === "") commonFit = fit;
      }

      var options =
        '<option value="none">None</option>' +
        '<option value="fit">Fit</option>' +
        '<option value="cover">Cover</option>' +
        '<option value="contain">Contain</option>';
      var searchKey = 'value="' + commonFit + '"';
      var index = options.indexOf(searchKey) + searchKey.length;
      return options.substr(0, index) + " selected" + options.substr(index);

      function getFitOption(node) {
        if (!elementUtilities.hasBackgroundImage(node)) return;

        var f = node.data("background-fit");
        var h = node.data("background-height");

        if (!f || !h) return;

        f = f.split(" ");
        h = h.split(" ");
        if (f[f.length - 1] === "none")
          return h[h.length - 1] === "auto" ? "none" : "fit";
        else return f[f.length - 1];
      }
    };

    elementUtilities.updateBackgroundImage = function (nodes, bgObj) {
      if (!nodes || nodes.length == 0 || !bgObj) return;

      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var obj = bgObj[node.data("id")];
        if (!obj || $.isEmptyObject(obj)) continue;

        var imgs = node.data("background-image")
          ? node.data("background-image").split(" ")
          : [];
        var xPos = node.data("background-position-x")
          ? node.data("background-position-x").split(" ")
          : [];
        var yPos = node.data("background-position-y")
          ? node.data("background-position-y").split(" ")
          : [];
        var widths = node.data("background-width")
          ? node.data("background-width").split(" ")
          : [];
        var heights = node.data("background-height")
          ? node.data("background-height").split(" ")
          : [];
        var fits = node.data("background-fit")
          ? node.data("background-fit").split(" ")
          : [];
        var opacities = node.data("background-image-opacity")
          ? ("" + node.data("background-image-opacity")).split(" ")
          : [];

        var index = -1;
        if (typeof obj["background-image"] === "string")
          index = imgs.indexOf(obj["background-image"]);
        else if (Array.isArray(obj["background-image"]))
          index = imgs.indexOf(obj["background-image"][0]);

        if (index < 0) continue;

        if (obj["background-image"] && imgs.length > index) {
          var tmp = imgs[index];
          imgs[index] = obj["background-image"];
          obj["background-image"] = tmp;
        }
        if (obj["background-fit"] && fits.length > index) {
          var tmp = fits[index];
          fits[index] = obj["background-fit"];
          obj["background-fit"] = tmp;
        }
        if (obj["background-width"] && widths.length > index) {
          var tmp = widths[index];
          widths[index] = obj["background-width"];
          obj["background-width"] = tmp;
        }
        if (obj["background-height"] && heights.length > index) {
          var tmp = heights[index];
          heights[index] = obj["background-height"];
          obj["background-height"] = tmp;
        }
        if (obj["background-position-x"] && xPos.length > index) {
          var tmp = xPos[index];
          xPos[index] = obj["background-position-x"];
          obj["background-position-x"] = tmp;
        }
        if (obj["background-position-y"] && yPos.length > index) {
          var tmp = yPos[index];
          yPos[index] = obj["background-position-y"];
          obj["background-position-y"] = tmp;
        }
        if (obj["background-image-opacity"] && opacities.length > index) {
          var tmp = opacities[index];
          opacities[index] = obj["background-image-opacity"];
          obj["background-image-opacity"] = tmp;
        }

        node.data("background-image", imgs.join(" "));
        node.data("background-position-x", xPos.join(" "));
        node.data("background-position-y", yPos.join(" "));
        node.data("background-width", widths.join(" "));
        node.data("background-height", heights.join(" "));
        node.data("background-fit", fits.join(" "));
        node.data("background-image-opacity", opacities.join(" "));
      }

      return bgObj;
    };

    elementUtilities.changeBackgroundImage = function (
      nodes,
      oldImg,
      newImg,
      firstTime,
      updateInfo,
      promptInvalidImage,
      validateURL
    ) {
      if (!nodes || nodes.length == 0 || !oldImg || !newImg) return;

      elementUtilities.removeBackgroundImage(nodes, oldImg);
      for (var key in newImg) {
        newImg[key]["firstTime"] = firstTime;
      }
      elementUtilities.addBackgroundImage(
        nodes,
        newImg,
        updateInfo,
        promptInvalidImage,
        validateURL
      );

      return {
        nodes: nodes,
        oldImg: newImg,
        newImg: oldImg,
        firstTime: false,
        promptInvalidImage: promptInvalidImage,
        validateURL: validateURL,
      };
    };

    // Add a background image to given nodes.
    elementUtilities.addBackgroundImage = function (
      nodes,
      bgObj,
      updateInfo,
      promptInvalidImage,
      validateURL
    ) {
      if (!nodes || nodes.length == 0 || !bgObj) return;

      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var obj = bgObj[node.data("id")];
        if (!obj || $.isEmptyObject(obj)) continue;

        // Load the image from local, else just put the URL
        if (obj["fromFile"]) loadBackgroundThenApply(node, obj);
        // Validity of given URL should be checked before applying it
        else if (obj["firstTime"]) {
          if (typeof validateURL === "function")
            validateURL(node, obj, applyBackground, promptInvalidImage);
          else checkGivenURL(node, obj);
        } else applyBackground(node, obj);
      }

      function loadBackgroundThenApply(node, bgObj) {
        var reader = new FileReader();
        var imgFile = bgObj["background-image"];

        // Check whether given file is an image file
        if (imgFile.type.indexOf("image") !== 0) {
          if (promptInvalidImage)
            promptInvalidImage("Invalid image file is given!");
          return;
        }

        reader.readAsDataURL(imgFile);

        reader.onload = function (e) {
          var img = reader.result;
          if (img) {
            bgObj["background-image"] = img;
            bgObj["fromFile"] = false;
            applyBackground(node, bgObj);
          } else {
            if (promptInvalidImage)
              promptInvalidImage("Given file could not be read!");
          }
        };
      }

      function checkGivenURL(node, bgObj) {
        var url = bgObj["background-image"];
        var extension = url.split(/[?#]/)[0].split(".").pop();
        var validExtensions = ["png", "svg", "jpg", "jpeg"];

        if (!validExtensions.includes(extension)) {
          if (typeof promptInvalidImage === "function")
            promptInvalidImage("Invalid URL is given!");
          return;
        }

        $.ajax({
          url: url,
          type: "GET",
          success: function (result, status, xhr) {
            applyBackground(node, bgObj);
          },
          error: function (xhr, status, error) {
            if (promptInvalidImage) promptInvalidImage("Invalid URL is given!");
          },
        });
      }

      function applyBackground(node, bgObj) {
        if (elementUtilities.hasBackgroundImage(node)) return;

        var imgs = node.data("background-image")
          ? node.data("background-image").split(" ")
          : [];
        var xPos = node.data("background-position-x")
          ? node.data("background-position-x").split(" ")
          : [];
        var yPos = node.data("background-position-y")
          ? node.data("background-position-y").split(" ")
          : [];
        var widths = node.data("background-width")
          ? node.data("background-width").split(" ")
          : [];
        var heights = node.data("background-height")
          ? node.data("background-height").split(" ")
          : [];
        var fits = node.data("background-fit")
          ? node.data("background-fit").split(" ")
          : [];
        var opacities = node.data("background-image-opacity")
          ? ("" + node.data("background-image-opacity")).split(" ")
          : [];

        var indexToInsert = imgs.length;

        // insert to length-1
        if (elementUtilities.hasCloneMarker(imgs)) {
          indexToInsert--;
        }

        imgs.splice(indexToInsert, 0, bgObj["background-image"]);
        fits.splice(indexToInsert, 0, bgObj["background-fit"]);
        opacities.splice(indexToInsert, 0, bgObj["background-image-opacity"]);
        xPos.splice(indexToInsert, 0, bgObj["background-position-x"]);
        yPos.splice(indexToInsert, 0, bgObj["background-position-y"]);
        widths.splice(indexToInsert, 0, bgObj["background-width"]);
        heights.splice(indexToInsert, 0, bgObj["background-height"]);

        node.data("background-image", imgs.join(" "));
        node.data("background-position-x", xPos.join(" "));
        node.data("background-position-y", yPos.join(" "));
        node.data("background-width", widths.join(" "));
        node.data("background-height", heights.join(" "));
        node.data("background-fit", fits.join(" "));
        node.data("background-image-opacity", opacities.join(" "));
        bgObj["firstTime"] = false;

        if (updateInfo) updateInfo();
      }
    };

    elementUtilities.hasCloneMarker = function (imgs) {
      var cloneImg =
        "data:image/svg+xml;utf8,%3Csvg%20width%3D%22100%22%20height%3D%22100%22%20viewBox%3D%220%200%20100%20100%22%20style%3D%22fill%3Anone%3Bstroke%3Ablack%3Bstroke-width%3A0%3B%22%20xmlns%3D%22http%3A//www.w3.org/2000/svg%22%20%3E%3Crect%20x%3D%220%22%20y%3D%220%22%20width%3D%22100%22%20height%3D%22100%22%20style%3D%22fill%3A%23838383%22/%3E%20%3C/svg%3E";
      return imgs.indexOf(cloneImg) > -1;
    };

    // Remove a background image from given nodes.
    elementUtilities.removeBackgroundImage = function (nodes, bgObj) {
      if (!nodes || nodes.length == 0 || !bgObj) return;

      for (var i = 0; i < nodes.length; i++) {
        var node = nodes[i];
        var obj = bgObj[node.data("id")];
        if (!obj) continue;

        var imgs = node.data("background-image")
          ? node.data("background-image").split(" ")
          : [];
        var xPos = node.data("background-position-x")
          ? node.data("background-position-x").split(" ")
          : [];
        var yPos = node.data("background-position-y")
          ? node.data("background-position-y").split(" ")
          : [];
        var widths = node.data("background-width")
          ? node.data("background-width").split(" ")
          : [];
        var heights = node.data("background-height")
          ? node.data("background-height").split(" ")
          : [];
        var fits = node.data("background-fit")
          ? node.data("background-fit").split(" ")
          : [];
        var opacities = node.data("background-image-opacity")
          ? ("" + node.data("background-image-opacity")).split(" ")
          : [];

        var index = -1;
        if (typeof obj["background-image"] === "string")
          index = imgs.indexOf(obj["background-image"].split(" ")[0]);
        else if (Array.isArray(obj["background-image"]))
          index = imgs.indexOf(obj["background-image"][0]);

        if (index > -1) {
          imgs.splice(index, 1);
          fits.splice(index, 1);
          opacities.splice(index, 1);
          xPos.splice(index, 1);
          yPos.splice(index, 1);
          widths.splice(index, 1);
          heights.splice(index, 1);
        }

        node.data("background-image", imgs.join(" "));
        node.data("background-position-x", xPos.join(" "));
        node.data("background-position-y", yPos.join(" "));
        node.data("background-width", widths.join(" "));
        node.data("background-height", heights.join(" "));
        node.data("background-fit", fits.join(" "));
        node.data("background-image-opacity", opacities.join(" "));
        bgObj["firstTime"] = false;
      }
    };

    elementUtilities.reverseEdge = function (edge) {
      var oldSource = edge.source().id();
      var oldTarget = edge.target().id();
      var oldPortSource = edge.data("portsource");
      var oldPortTarget = edge.data("porttarget");
      var segmentPoints = edge.segmentPoints();
      var controlPoints = edge.controlPoints();

      edge.data().source = oldTarget;
      edge.data().target = oldSource;
      edge.data().portsource = oldPortTarget;
      edge.data().porttarget = oldPortSource;
      edge = edge.move({
        target: oldSource,
        source: oldTarget,
      });

      if (Array.isArray(segmentPoints)) {
        segmentPoints.reverse();
        edge.data().bendPointPositions = segmentPoints;
        if (Array.isArray(controlPoints)) {
          controlPoints.reverse();
          edge.data().controlPointPositions = controlPoints;
        }
        var edgeEditing = cy.edgeEditing("get");
        edgeEditing.initAnchorPoints(edge);
      }

      return edge;
    };
  }

  return elementUtilitiesExtender;
};
